from datetime import datetime, timedelta

import allure
import pytest

from refinitiv.data.content.ipa.dates_and_calendars import holidays
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.helpers import (
    check_response_status,
    get_async_response_from_definitions,
    check_dataframe_column_date_for_datetime_type,
    check_non_empty_response_data,
    check_response_dataframe_contains_columns_names,
    check_extended_params_were_sent_in_request,
)

EXPECTED_COLUMN_NAMES = [
    "name",
    "calendars",
    "countries",
    "tag",
    "date",
]


first_definition = holidays.Definition(
    tag="my request",
    start_date=datetime(2020, 5, 2),
    end_date=timedelta(-30),
    calendars=["UKR", "FRA"],
    currencies=["EUR"],
    holiday_outputs=["Date", "Names", "Calendars", "Countries"],
)

second_definition = holidays.Definition(
    tag="my second request",
    start_date="2018-12-31",
    end_date=datetime.now(),
    calendars=["UKR", "FRA"],
    currencies=["EUR"],
    holiday_outputs=["Date", "Names", "Calendars", "Countries"],
)

invalid_definition = holidays.Definition(
    tag="my second request",
    start_date=datetime.now(),
    end_date=timedelta(-30),
    calendars=["UKR", "FRA"],
    currencies=["EUR"],
    holiday_outputs=["Date", "Names", "Calendars", "Countries"],
)


@allure.suite("Dates And Calendars")
@allure.feature("Content object - Holidays")
@allure.severity(allure.severity_level.CRITICAL)
class TestHolidays:
    @allure.title("Create a holidays definition with valid params")
    @pytest.mark.caseid("C46153589")
    @pytest.mark.parametrize(
        "start_date, end_date, calendars, currencies, holiday_outputs",
        [
            (
                "2018-12-31",
                "2019-01-03",
                ["UKR", "FRA"],
                ["EUR"],
                ["Date", "Names", "Calendars", "Countries"],
            ),
            (
                timedelta(-180),
                timedelta(0),
                ["UKR", "FRA"],
                None,
                [holidays.HolidayOutputs.DATE, holidays.HolidayOutputs.NAMES],
            ),
            (
                timedelta(-180),
                datetime.now(),
                None,
                ["USD"],
                ["Date", "Calendars", "Names"],
            ),
            (
                datetime(2019, 4, 13),
                datetime.now(),
                None,
                ["USD"],
                None,
            ),
        ],
    )
    def test_holidays_definition_with_valid_params(
        self, open_session, start_date, end_date, calendars, currencies, holiday_outputs
    ):
        response = holidays.Definition(
            tag="my request 2",
            start_date=start_date,
            end_date=end_date,
            calendars=calendars,
            currencies=currencies,
            holiday_outputs=holiday_outputs,
        ).get_data()

        assert isinstance(response.data.holidays, list)

        check_response_status(response, HttpStatusCode.TWO_HUNDRED, HttpReason.OK)
        check_non_empty_response_data(response)

        if holiday_outputs is not None:
            check_response_dataframe_contains_columns_names(
                response, EXPECTED_COLUMN_NAMES
            )
            check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Create a holidays definition with valid params asynchronously")
    @pytest.mark.caseid("C46323263")
    async def test_holidays_definition_with_valid_params_async(self, open_session_async):
        valid_response, invalid_response = await get_async_response_from_definitions(
            first_definition, invalid_definition
        )

        check_non_empty_response_data(valid_response)
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="IPA_library.ErrorBusinessCalendar_EndDateMustBeGreaterThanStartDate",
            expected_error_message="Business Calendar: end date must be greater than start date.",
        )

    @allure.title("Create a holidays definitions with valid and invalid params")
    @pytest.mark.caseid("C46153592")
    def test_holidays_definitions_with_valid_and_invalid_params(self, open_session):
        response = holidays.Definitions(
            [first_definition, second_definition, invalid_definition]
        ).get_data()

        check_non_empty_response_data(response)
        check_response_status(
            response=response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="IPA_library.ErrorBusinessCalendar_EndDateMustBeGreaterThanStartDate",
            expected_error_message="Business Calendar: end date must be greater than start date.",
        )
        check_response_dataframe_contains_columns_names(response, EXPECTED_COLUMN_NAMES)
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title(
        "Create a holidays definition with invalid params and getting RDError"
    )
    @pytest.mark.parametrize(
        "expected_error",
        [
            (
                "Error code 400 | Invalid input: at least one currency code or calendar code must be specified."
            )
        ],
    )
    @pytest.mark.caseid("C46153599")
    def test_holidays_definition_with_invalid_params(
        self, open_session, expected_error
    ):
        with pytest.raises(RDError) as error:
            holidays.Definition(
                start_date=timedelta(-30), end_date=timedelta(0)
            ).get_data()
        assert str(error.value) == expected_error

    @allure.title(
        "Create a holidays definitions with invalid params and getting RDError"
    )
    @pytest.mark.parametrize(
        "expected_error",
        [
            (
                "Error code IPA_library.ErrorBusinessCalendar_EndDateMustBeGreaterThanStartDate | Business Calendar: "
                "end date must be greater than start date."
            )
        ],
    )
    @pytest.mark.caseid("C46153595")
    def test_holidays_definitions_with_invalid_params(
        self, open_session, expected_error
    ):
        with pytest.raises(RDError) as error:
            holidays.Definitions(invalid_definition).get_data()
        assert str(error.value) == expected_error

    @allure.title("Create a holidays definitions with valid params asynchronously")
    @pytest.mark.caseid("C46323265")
    async def test_holidays_definitions_with_valid_params_async(self, open_session_async):
        valid_response, invalid_response = await get_async_response_from_definitions(
            holidays.Definitions([first_definition, second_definition]),
            holidays.Definitions(invalid_definition),
        )

        check_non_empty_response_data(valid_response)
        check_response_status(
            response=valid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="IPA_library.ErrorBusinessCalendar_EndDateMustBeGreaterThanStartDate",
            expected_error_message="Business Calendar: end date must be greater than start date.",
        )

    @allure.title("Create Holidays definition with extended params field and get data")
    @pytest.mark.parametrize(
        "extended_param",
        [
            {
                "tag": "new tag",
                "startDate": "2022-07-04",
                "holidayOutputs": ["Date"],
            }
        ],
    )
    @pytest.mark.caseid("C46153590")
    def test_holidays_definitions_with_extended_params(
        self, open_session, extended_param
    ):
        response = holidays.Definition(
            tag="my request",
            start_date=timedelta(-360),
            end_date=datetime.now(),
            calendars=["UKR", "FRA"],
            currencies=["EUR"],
            holiday_outputs=["Date", "Names", "Calendars", "Countries"],
            extended_params=extended_param,
        ).get_data()

        check_extended_params_were_sent_in_request(response, extended_param)
        check_non_empty_response_data(response)
        check_response_status(
            response=response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
        )



from inspect import signature

import numpy as np
import pytest

from refinitiv.data.content.ipa._curves._models._instrument import Instrument
from refinitiv.data.content.ipa._curves._forward_curve_definition import (
    ForwardCurveDefinition,
)
from refinitiv.data.content.ipa._curves._forward_curve_request_item import (
    ForwardCurveRequestItem,
)
from refinitiv.data.content.ipa._curves._models import CurvePoint
from refinitiv.data.content.ipa._curves._models._curve import (
    ZcCurve,
    ForwardCurve,
    Curve,
)
from refinitiv.data.content.ipa._curves._swap_zc_curve_definition import (
    SwapZcCurveDefinition,
)
from refinitiv.data.content.ipa._curves._swap_zc_curve_parameters import (
    SwapZcCurveParameters,
)
from refinitiv.data.content.ipa._enums import Axis
from refinitiv.data.content.ipa import curves as public_curves
from refinitiv.data.content.ipa.curves import (
    zc_curves,
    forward_curves,
)
from tests.unit.conftest import (
    remove_dunder_methods,
    remove_private_attributes,
    has_property_names_in_class,
    get_property_names,
    StubSession,
    StubResponse,
)
from tests.unit.content.ipa._curves.conftest import (
    create_swap_zc_curve_def,
    create_forward_curve_def,
    zc_curves_definitions_json,
)


def test_surfaces_version():
    # given
    expected = "1.0.59"

    # when
    from refinitiv.data.content.ipa import _curves

    testing_version = _curves.__version__

    # then
    assert testing_version == expected, testing_version


def test_zc_curve():
    # given
    kwargs = {"isDiscountCurve": True}
    curve = ZcCurve(x=np.array([]), y=np.array([]), index_tenor="index_tenor", **kwargs)

    # when
    s = str(curve)

    # then
    assert s


def test_forward_curve():
    # given
    kwargs = {
        "forwardCurveTag": "forward_curve_tag",
        "forwardStart": "forward_start",
        "indexTenor": "index_tenor",
    }
    curve = ForwardCurve(x=np.array([]), y=np.array([]), **kwargs)

    # then
    assert curve.x is not None and curve.y is not None


def test_get_axis_removed():
    """get_axis removed in favor of direct access to x, y attributes
    https://jira.refinitiv.com/browse/EAPI-2687
    """
    curve = ForwardCurve(x=np.array([]), y=np.array([]))

    with pytest.raises(AttributeError):
        curve.get_axis()


@pytest.mark.skip(reason="Scipy 1.9.x has different float result")
def test_curve_get_point_by_axis_x():
    # given
    # array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    x = np.arange(0, 10)
    # array([1.        , 0.71653131, 0.51341712, 0.36787944, 0.26359714,
    #        0.1888756 , 0.13533528, 0.09697197, 0.06948345, 0.04978707])
    y = np.exp(-x / 3.0)
    curve = Curve(x=x, y=y)

    # when
    point = curve.get_point(5.5, Axis.X)

    # then
    assert point.x == 5.5 and point.y == 0.16210544303708727, point


@pytest.mark.skip(reason="Scipy 1.9.x has different float result")
def test_curve_get_point_by_axis_y():
    # given
    # array([0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    x = np.arange(0, 10)
    # array([1.        , 0.71653131, 0.51341712, 0.36787944, 0.26359714,
    #        0.1888756 , 0.13533528, 0.09697197, 0.06948345, 0.04978707])
    y = np.exp(-x / 3.0)
    curve = Curve(x=x, y=y)

    # when
    point = curve.get_point(0.5, Axis.Y)

    # then
    assert point.x == 2.0921900035081817 and point.y == 0.5, point


def test_curve_point_to_string_using_str_dunder_method():
    # given
    point = CurvePoint(0, 0)

    # when
    s = str(point)

    # then
    assert s


def test_curve_point_to_string_using_repr_dunder_method():
    # given
    point = CurvePoint(0, 0)

    # when
    s = repr(point)

    # then
    assert s


def test_curve_get_point_raise_error_if_axis_is_invalid():
    # given
    curve = Curve(x=np.array([]), y=np.array([]))

    # then
    with pytest.raises(ValueError):
        # when
        curve.get_point(0.5, Axis.Z)


def test_parameters_forward_curve_definition():
    """
    ipa.forward_curves.ForwardCurveDefinition(
        index_tenor='3M',  # required
        forward_curve_tag='ForwardTag',  # required
        forward_start_date='2021-02-01',  # optional
        forward_curve_tenors=['0D', '1D'],  # optional
        forward_start_tenor='some_start_tenor'  # optional
        )
    """
    try:
        forward_curves.ForwardCurveDefinition(
            index_tenor="3M",
            forward_curve_tag="ForwardTag",
            forward_start_date="2021-02-01",
            forward_curve_tenors=["0D", "1D"],
            forward_start_tenor="some_start_tenor",
        )
    except Exception as e:
        assert False, e


def test_parameters_swap_zc_curve_definition():
    """
    ipa.forward_curves.SwapZcCurveDefinition(
        currency='EUR',  # all fields are optional
        indexName='EURIBOR',
        discountingTenor='OIS',
        )
    """
    try:
        forward_curves.SwapZcCurveDefinition(
            currency="EUR",
            index_name="EURIBOR",
            discounting_tenor="OIS",
        )
    except Exception as e:
        assert False, e


def test_parameters_swap_zc_curve_definition_all_optional():
    try:
        forward_curves.SwapZcCurveDefinition()
    except Exception as e:
        assert False, e


def test_definitions_single_forward_curve_full_definition():
    """
    ipa.forward_curves.Definition(
        curve_definition=swap_sc_curve_def,  # required
        forward_curve_definitions=[forward_curve_def],  # required
        curveParameters=...,  # optional (SwapZcCurveParameters),
        curveTag='some_tag'  # optional (string)
        )
    """
    try:
        forward_curves.Definition(
            curve_definition=create_swap_zc_curve_def(),
            forward_curve_definitions=[create_forward_curve_def()],
            curve_parameters=forward_curves.SwapZcCurveParameters(),
            curve_tag="some_tag",
        )
    except Exception as e:
        assert False, e


def test_forward_curves_scope():
    expected_attributes = [
        "AssetClass",
        "CalendarAdjustment",
        "CompoundingType",
        "ConvexityAdjustment",
        "DayCountBasis",
        "Definition",
        "Definitions",
        "ExtrapolationMode",
        "ForwardCurveDefinition",
        "InterpolationMode",
        "Outputs",
        "PriceSide",
        "RiskType",
        "Step",
        "SwapZcCurveDefinition",
        "SwapZcCurveParameters",
        "Turn",
    ]
    testing_attributes = dir(forward_curves)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes, set(expected_attributes) - set(
        testing_attributes
    )


def test_zc_curves_scope():
    expected_attributes = [
        "AssetClass",
        "CalendarAdjustment",
        "CompoundingType",
        "DayCountBasis",
        "Definition",
        "Definitions",
        "ExtrapolationMode",
        "MarketDataAccessDeniedFallback",
        "Outputs",
        "PriceSide",
        "RiskType",
        "ZcCurveDefinitions",
        "ZcCurveParameters",
        "ZcInterpolationMode",
    ]
    testing_attributes = dir(zc_curves)
    testing_attributes = remove_dunder_methods(testing_attributes)
    testing_attributes = remove_private_attributes(testing_attributes)
    assert expected_attributes == testing_attributes


@pytest.mark.parametrize(
    argnames="input_data",
    ids=[
        "ForwardCurveRequestItem",
        "ForwardCurveDefinition",
        "SwapZcCurveDefinition",
        "Instrument",
    ],
    argvalues=[
        (
            ForwardCurveRequestItem,
            {
                "curve_definition": SwapZcCurveDefinition(
                    currency="EUR",
                    index_name="EURIBOR",
                    discounting_tenor="OIS",
                ),
                "forward_curve_definitions": [ForwardCurveDefinition(index_tenor="3M")],
                "curve_parameters": SwapZcCurveParameters(),
                "curve_tag": "curve_tag",
            },
        ),
        (
            ForwardCurveDefinition,
            {
                "index_tenor": "3M",
                "forward_curve_tag": "ForwardTag",
                "forward_start_date": "2021-02-01",
                "forward_curve_tenors": ["0D", "1D"],
                "forward_start_tenor": "some_start_tenor",
            },
        ),
        (
            SwapZcCurveDefinition,
            {
                "currency": "EUR",
                "index_name": "EURIBOR",
                "index_tenors": ["3M"],
                "discounting_tenor": "OIS",
                "main_constituent_asset_class": forward_curves.AssetClass.SWAP,
                "risk_type": forward_curves.RiskType.INTEREST_RATE,
                "market_data_location": "market_data_location",
                "name": "name",
                "source": "Refinitiv",
                "id": "id",
            },
        ),
        (
            Instrument,
            {
                "instrument_code": "EUR",
                "value": "EURIBOR",
            },
        ),
    ],
)
def test_forward_curve_request_item_parameter(input_data):
    cls, kwargs = input_data
    args_names = list(kwargs.keys())
    inst = cls(**kwargs)

    s = signature(cls.__init__)
    assert len(s.parameters) == (len(args_names) + 1)  # +1 for (self)

    assert has_property_names_in_class(cls, args_names), set(args_names) - set(
        get_property_names(cls)
    )

    for k, v in kwargs.items():
        attr = getattr(inst, k)
        assert attr == v, k


@pytest.mark.parametrize(
    "esg_definition, expected_repr",
    [
        (
            public_curves.zc_curves.ZcCurveDefinitions(),
            "<refinitiv.data.content.ipa.curves.zc_curves.ZcCurveDefinitions object at {0}>",
        ),
        (
            public_curves.zc_curves.ZcCurveParameters(),
            "<refinitiv.data.content.ipa.curves.zc_curves.ZcCurveParameters object at {0}>",
        ),
        (
            public_curves.zc_curves.Definitions(public_curves.zc_curves.Definition()),
            "<refinitiv.data.content.ipa.curves.zc_curves.Definitions object at {0}>",
        ),
        (
            public_curves.zc_curves.Definition(),
            "<refinitiv.data.content.ipa.curves.zc_curves.Definition object at {0}>",
        ),
        (
            public_curves.forward_curves.SwapZcCurveParameters(),
            "<refinitiv.data.content.ipa.curves.forward_curves.SwapZcCurveParameters object at {0}>",
        ),
        (
            public_curves.forward_curves.SwapZcCurveDefinition(),
            "<refinitiv.data.content.ipa.curves.forward_curves.SwapZcCurveDefinition object at {0}>",
        ),
        (
            public_curves.forward_curves.Definitions(
                public_curves.forward_curves.Definition()
            ),
            "<refinitiv.data.content.ipa.curves.forward_curves.Definitions object at {0}>",
        ),
        (
            public_curves.forward_curves.Definition(),
            "<refinitiv.data.content.ipa.curves.forward_curves.Definition object at {0}>",
        ),
    ],
)
def test_esg_content_repr(esg_definition, expected_repr):
    # given
    obj_id = hex(id(esg_definition))

    # when
    s = repr(esg_definition)

    # then
    assert s == expected_repr.format(obj_id)


def test_zc_curves_definition_attribute_error():
    # given
    response = StubResponse(zc_curves_definitions_json)
    session = StubSession(is_open=True, response=response)
    zc_curve_definition = zc_curves.Definition(
        curve_definition=zc_curves.ZcCurveDefinitions(
            currency="CHF",
            name="CHF LIBOR Swap ZC Curve",
            discounting_tenor="OIS",
        ),
        curve_parameters=zc_curves.ZcCurveParameters(
            valuation_date="2019-08-21", use_steps=True
        ),
    )
    definition = zc_curves.Definitions([zc_curve_definition, zc_curves.Definition()])
    response = definition.get_data(session)

    try:
        # when
        response.data.curves
    except Exception as e:
        assert False, str(e)
    else:
        # then
        assert True



from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    FxForwardTurnFields,
)


def test_fx_forward_turn_fields():
    # given
    expected_dict = {"bid": 10.0, "ask": 10.0}

    # when
    testing_obj = FxForwardTurnFields(
        bid=10.0,
        ask=10.0,
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.bid == 10.0
    assert testing_obj.ask == 10.0



import abc
from typing import Dict, Optional

import pandas as pd
from pandas import DataFrame

from ._context import Context
from ..._tools import ohlc
from ..._tools._dataframe import (
    convert_dtypes,
)


class CustInstContext(Context, abc.ABC):
    @property
    def can_get_data(self) -> bool:
        return bool(self.universe.cust_inst)

    @property
    def can_build_df(self) -> bool:
        return bool(self._cust_inst_data and not (self._adc_data or self._hp_data))

    @property
    def can_join_hp_multiindex_df(self) -> bool:
        return bool(
            not self._adc_data
            and (self._hp_data and self._cust_inst_data and len(self.universe.hp) > 1)
        )

    @property
    def raw(self) -> Optional[Dict]:
        return self._cust_inst_data and self._cust_inst_data.raw

    @property
    def df(self) -> Optional[DataFrame]:
        return self._cust_inst_data and self._cust_inst_data.df

    def build_df(self, use_field_names_in_headers: bool, *args) -> DataFrame:
        """
        Builds custom instrument dataframe.

        Parameters
        ----------
        use_field_names_in_headers : bool
            Flag to use field names in headers
        args
            Variable length additional arguments list.

        Returns
        -------
        pd.DataFrame
            Custom instrument dataframe.

        """
        fields = self.fields
        df = self.df

        if fields:
            data = self.prepare_data(self.raw, fields)
            headers = self.prepare_headers(fields.raw)
            use_multiindex = len(fields.raw) > 1 and len(self.universe.cust_inst) > 1
            df = self._build_common_df(
                data, headers, use_field_names_in_headers, use_multiindex=use_multiindex
            )

        df = convert_dtypes(df)
        df.ohlc = ohlc.__get__(df, None)
        return df

    def join_hp_multiindex_df(self, hp_df: DataFrame, use_field_names_in_headers: bool):
        """
        Joins historical pricing multiindex dataframe with custom instruments dataframe.

        Parameters
        ----------
        hp_df : pd.DataFrame
            Historical pricing multiindex dataframe.
        use_field_names_in_headers : bool
            Flag to use fields names in headers.

        Returns
        -------
        pd.DataFrame
            Historical pricing multiindex dataframe, joined with custom instruments
            dataframe.
        """
        fields = self._get_fields_from_raw()
        data = self.prepare_data(self.raw, fields)
        headers = self.prepare_headers(fields)
        cust_df = self._build_common_df(
            data, headers, use_field_names_in_headers, use_multiindex=True
        )
        joined_df = hp_df.join(cust_df, how="outer")
        df = convert_dtypes(joined_df)
        df.ohlc = ohlc.__get__(df, None)
        return df

    def _build_common_df(
        self,
        data: list,
        headers: list,
        use_field_names_in_headers: bool,
        use_multiindex: bool,
    ):
        """
        Builds common dataframe for custom instruments data.

        Parameters
        ----------
        data : list
            Data to create dataframe.
        headers : list
            Headers to create dataframe.
        use_field_names_in_headers : bool
            Flag to use fields names in headers.
        use_multiindex : bool
            Flag to use multiindex for dataframe creation or not.

        Returns
        -------
        DataFrame
        """
        df = self.dfbuilder.build_date_as_index(
            {"data": data, "headers": headers},
            use_field_names_in_headers,
            use_multiindex=use_multiindex,
        )
        return df

    def _get_fields_from_raw(self):
        fields = []
        if isinstance(self.raw, list):
            headers = self.raw[0]["headers"]

        else:
            headers = self.raw["headers"]

        for header in headers:
            name = header.get("name")
            if name and name.lower() not in {"date", "instrument"}:
                fields.append(name)

        return fields

    def _get_fields_from_headers(self, headers, use_field_names_in_headers):
        name = "name" if use_field_names_in_headers else "title"
        return [
            header[name]
            for header in self.dfbuilder.get_headers({"headers": headers})
            if header[name].lower() not in {"date", "instrument"}
        ]

    def prepare_to_build(
        self, use_field_names_in_headers, df: DataFrame, headers, *args
    ):
        """
        Creates dataframe with ADC, historical pricing and custom instruments data.

        Join or merge previously created ADC and historical pricing dataframe with
        custom instruments dataframe.

        Parameters
        ----------
        use_field_names_in_headers : bool
            Use
        df : pd.DataFrame
            Previously created ADC and historical pricing dataframe.
        headers : List
            Common headers for building dataframe
        args
            Variable length additional arguments list.

        Returns
        -------
        pd.Dataframe that includes ADC, hp and custom instruments data.

        """
        if not self.fields:
            fields = self._get_fields_from_raw()

        else:
            fields = self._get_fields_from_headers(headers, use_field_names_in_headers)

        data = self.prepare_data(self.raw, fields)
        headers = self.prepare_headers(fields)
        cust_df = self._build_common_df(
            data, headers, use_field_names_in_headers, use_multiindex=True
        )

        if (not self._adc_data and self._hp_data) or (
            self._adc_data and not self._hp_data
        ):
            df = df.join(cust_df, how="outer")
        else:
            df = pd.merge(df, cust_df, on=["Date"])

        self._cust_inst_data._df = df



import pytest

from refinitiv.data.content.pricing._stream_facade import Stream
from refinitiv.data import OpenState
from tests.unit.conftest import StubSession


def test_pricing_stream_raise_error_if_no_session():
    # then
    with pytest.raises(AttributeError):
        # when
        stream = Stream()


def test_pricing_stream_raise_error_if_bad_universe():
    # given
    session = StubSession(is_open=True)
    stream = Stream(session=session, universe=[None])

    # then
    with pytest.raises(ValueError):
        # when
        stream.open()


def test_pricing_stream__repr__():
    # given
    session = StubSession(is_open=True)
    stream = Stream(session=session, universe="universe")
    hex_id = hex(id(stream))
    expected_value = f"<refinitiv.data.content.pricing.Stream object at {hex_id} {{name='universe'}}>"

    # when
    testing_value = repr(stream)

    # then
    assert testing_value == expected_value


def test_pricing_stream__iter__(pricing_stream):
    # when
    iterator = iter(pricing_stream)

    # then
    with pytest.raises(StopIteration):
        next(iterator)


def test_pricing_stream__getitem__(pricing_stream):
    # when
    testing_value = pricing_stream["item"]

    # then
    assert testing_value == {}


def test_pricing_stream__len__(pricing_stream):
    # when
    testing_value = len(pricing_stream)

    # then
    assert testing_value == 0


def test_pricing_stream_close(pricing_stream):
    # when
    state = pricing_stream.close()

    # then
    assert state is OpenState.Closed, state


def test_pricing_stream_get_snapshot(pricing_stream):
    # when
    snapshot = pricing_stream.get_snapshot()

    # then
    assert snapshot.empty is True, snapshot


def test_pricing_stream_on_refresh(pricing_stream):
    # when
    testing_value = pricing_stream.on_refresh(lambda: None)

    # then
    assert testing_value is pricing_stream


def test_pricing_stream_on_update(pricing_stream):
    # when
    testing_value = pricing_stream.on_update(lambda: None)

    # then
    assert testing_value is pricing_stream


def test_pricing_stream_on_status(pricing_stream):
    # when
    testing_value = pricing_stream.on_status(lambda: None)

    # then
    assert testing_value is pricing_stream


def test_pricing_stream_on_complete(pricing_stream):
    # when
    testing_value = pricing_stream.on_complete(lambda: None)

    # then
    assert testing_value is pricing_stream


def test_pricing_stream_on_error(pricing_stream):
    # when
    testing_value = pricing_stream.on_error(lambda: None)

    # then
    assert testing_value is pricing_stream



from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    ValuationTime,
)


def test_valuation_time():
    # given
    expected_dict = {
        "cityName": "string",
        "localTime": "string",
        "marketIdentifierCode": "string",
        "timeZoneOffset": "string",
    }

    # when
    testing_obj = ValuationTime(
        city_name="string",
        local_time="string",
        market_identifier_code="string",
        time_zone_offset="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.city_name == "string"
    assert testing_obj.local_time == "string"
    assert testing_obj.market_identifier_code == "string"
    assert testing_obj.time_zone_offset == "string"



from ._context import RDPMixin, UDFMixin
from ._context_factory import get_hp_context, get_adc_context, get_cust_inst_context
from ._adc_context import ADCContext
from ._cust_inst_context import CustInstContext
from ._hp_context import HPContext



import allure
import pytest

from refinitiv.data.content.ipa.curves import forward_curves
from refinitiv.data.errors import RDError
from tests.integration.constants_list import HttpStatusCode, HttpReason
from tests.integration.content.ipa.curves.conftest import (
    forward_curve_definition_01,
    forward_curve_definition_02,
    forward_curve_definition_03,
    check_curves,
    check_curve_response,
)
from tests.integration.helpers import (
    get_async_response_from_definitions,
    check_response_status,
    assert_error,
    check_dataframe_column_date_for_datetime_type,
)


def forward_curves_definition_01():
    definition = forward_curves.Definitions(
        [forward_curve_definition_02(), forward_curve_definition_01()]
    )

    return definition


def forward_curves_definition_02():
    definition = forward_curves.Definitions(
        [forward_curve_definition_01(), forward_curve_definition_03()]
    )

    return definition


@allure.suite("Content object - ForwardCurves")
@allure.feature("Content object - ForwardCurves")
@allure.severity(allure.severity_level.CRITICAL)
class TestForwardCurves:
    @allure.title(
        "Create a forward curves definition object with valid params and get data"
    )
    @pytest.mark.caseid("37925242")
    @pytest.mark.smoke
    def test_create_forward_curves_definition_with_valid_params(self, open_session):
        response = forward_curves_definition_01().get_data()

        check_curve_response(response, expected_currency="EUR")
        check_dataframe_column_date_for_datetime_type(response)

    @allure.title("Create a empty forward curves definition object and get data")
    @pytest.mark.caseid("37925244")
    def test_create_empty_forward_curves_definition(self):
        with pytest.raises(TypeError) as error:
            forward_curves.Definitions().get_data()

        assert_error(error, "universe")

    @allure.title("Create an invalid forward curves definition object and get data")
    @pytest.mark.caseid("39592733")
    def test_create_invalid_forward_curves_definition(self, open_session):
        with pytest.raises(
            RDError,
            match="Error code QPS-Curves.6 | Invalid input: curveDefinition is missing",
        ):
            forward_curves.Definitions([forward_curves.Definition()]).get_data()

    @allure.title(
        "Create a forward curves definition object and check curve parameters"
    )
    @pytest.mark.caseid("37925245")
    def test_create_forward_curves_definition_and_check_curve_parameters(
        self, open_session
    ):
        response = forward_curves_definition_01().get_data()

        check_curves(response)

    @allure.title("Create a forward curves definition object async")
    @pytest.mark.caseid("37925246")
    async def test_create_forward_curves_definition_async(self, open_session_async):
        invalid_response, valid_response = await get_async_response_from_definitions(
            forward_curves.Definitions([forward_curves.Definition()]),
            forward_curves_definition_01(),
        )

        check_response_status(
            response=invalid_response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="QPS-Curves.6",
            expected_error_message="Invalid input: curveDefinition is missing",
        )
        check_curves(valid_response)
        check_curve_response(valid_response, expected_currency="EUR")

    @allure.title(
        "Create a forward curves definition object with valid and invalid forward curve definition and get data"
    )
    @pytest.mark.caseid("39592730")
    @pytest.mark.smoke
    def test_create_forward_curves_definition_with_valid_and_invalid_forward_curve(
        self, open_session
    ):
        response = forward_curves_definition_02().get_data()

        check_curve_response(response, expected_currency="EUR")
        check_response_status(
            response=response,
            expected_status_code=HttpStatusCode.TWO_HUNDRED,
            expected_http_reason=HttpReason.OK,
            expected_error_code="QPS-Curves.6",
            expected_error_message="Invalid input: curveDefinition is missing",
        )
        check_curves(response)



from ..._enums import (
    AdjustInterestToPaymentDate,
    AmortizationFrequency,
    AmortizationType,
    BenchmarkYieldSelectionMode,
    BusinessDayConvention,
    CreditSpreadType,
    DateRollingConvention,
    DayCountBasis,
    Direction,
    DividendType,
    Frequency,
    IndexAverageMethod,
    IndexCompoundingMethod,
    InflationMode,
    InterestType,
    PriceSide,
    ProjectedIndexCalculationMethod,
    QuoteFallbackLogic,
    RedemptionDateType,
    Rounding,
    RoundingType,
    StubRule,
    VolatilityTermStructureType,
    VolatilityType,
    YieldType,
)



from refinitiv.data import _configure as configure


def test_data_grid_underlying_platform_default_parameter():
    # given
    expected = "udf"
    actual = configure.get_param("apis.data.datagrid.underlying-platform")

    # then
    assert actual == expected



import pandas as pd

from refinitiv.data._tools._dataframe import (
    convert_df_columns_to_datetime,
    convert_str_to_datetime,
)
from .conftest import TEST_CASES
from ..conftest import parametrize_with_test_case


@parametrize_with_test_case(
    "df,pattern,utc,delete_tz,expected_dtype,expected_df",
    TEST_CASES,
    "Simple DataFrame not valid datetime",
    "Simple DataFrame without timezone",
    "Simple DataFrame with timezone",
    "Simple DataFrame without timezone with flag utc=True",
    "Simple DataFrame with timezone with flag utc=True",
    "Simple DataFrame with timezone with flag utc=None, delete_tz=True",
    "Simple DataFrame without timezone with flag utc=True, delete_tz=True",
    "Simple DataFrame with timezone with flag utc=True, delete_tz=True",
)
def test_convert_df_columns_to_datetime(
    df, pattern, utc, delete_tz, expected_dtype, expected_df
):
    # when
    testing_df = convert_df_columns_to_datetime(df, pattern, utc, delete_tz)

    # then
    assert testing_df.Date.dtype == expected_dtype

    try:
        diff_df = testing_df.compare(expected_df)
    except ValueError as e:
        assert False, str(e)
    else:
        assert diff_df.empty is True, diff_df.to_string()


def test_not_valid_date_for_convert_str_to_datetime():
    # when
    testing = convert_str_to_datetime("1207")

    # then
    assert testing is pd.NaT



import numpy as np
import pandas as pd
import pytest

from refinitiv.data._content_type import ContentType
from refinitiv.data.content._df_build_type import DFBuildType
from refinitiv.data.content._df_builder_factory import get_dfbuilder
from tests.unit.content.data_for_tests import (
    NONE_REPLACED_WITH_NA_RDP_DATA,
    NONE_REPLACED_WITH_NA_UDF_DATA,
    NAN_REPLACED_WITH_NA_RDP_DATA,
    NAN_REPLACED_WITH_NA_UDF_DATA,
    EMPTY_STRING_LEAVE_AS_IT_IS_RDP_DATA,
    EMPTY_STRING_LEAVE_AS_IT_IS_UDF_DATA,
    INSTRUMENT_AND_CURRENCY_ARE_EMPTY_STRINGS_RDP_DATA,
    INSTRUMENT_AND_CURRENCY_ARE_EMPTY_STRINGS_UDF_DATA,
    TWO_SAME_DATES_DOES_NOT_MERGE_IN_ONE_RDP_DATA,
    TWO_SAME_DATES_DOES_NOT_MERGE_IN_ONE_UDF_DATA,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA,
    DELETE_ROW_IF_DATE_IS_NONE_UDF_DATA,
    DELETE_ROW_IF_DATE_IS_NONE_RDP_DATA,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA_4,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA_4,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA_3,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA_3,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA_2,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA_2,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA_1,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA_1,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA_5,
    MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA_5,
    TWO_UNIVERSES_ONE_FIELD_RDP_DATA,
    TWO_UNIVERSES_ONE_FIELD_UDF_DATA,
)

"""
https://confluence.refinitiv.com/display/EF/RD+Lib+Python+-+Technical+Requirements+-+DataFrame+Requirements
"""


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, NONE_REPLACED_WITH_NA_RDP_DATA),
        (ContentType.DATA_GRID_UDF, NONE_REPLACED_WITH_NA_UDF_DATA),
    ],
    ids=["None replaced with NA. RDP", "None replaced with NA. UDF"],
)
def test_none_replaced_with_na(content_type, input_data, dfbuild_type):
    # when
    dataframe_build = get_dfbuilder(content_type, dfbuild_type)
    """
    #### INDEX ####
    
      Instrument       Date Currency
    0     GOOG.O 2020-01-20     <NA>
    1     GOOG.O 2020-12-31     <NA>
    
    #### DATE_AS_INDEX #### 
    
    GOOG.O     Currency
    Date               
    2020-12-31     <NA>
    2020-01-20     <NA>
    """
    testing_df = dataframe_build(input_data)

    # then
    assert testing_df["Currency"][0] is pd.NA
    assert testing_df["Currency"][1] is pd.NA


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, NAN_REPLACED_WITH_NA_RDP_DATA),
        (ContentType.DATA_GRID_UDF, NAN_REPLACED_WITH_NA_UDF_DATA),
    ],
    ids=["NaN replaced with NA. RDP", "NaN replaced with NA. UDF"],
)
def test_nan_replaced_with_na(content_type, input_data, dfbuild_type):
    # when
    dataframe_build = get_dfbuilder(content_type, dfbuild_type)
    """
    #### INDEX ####
    
      Instrument       Date Currency
    0     GOOG.O 2020-01-20     1000
    1     GOOG.O 2020-12-31     <NA>

    #### DATE_AS_INDEX ####
     
    GOOG.O     Currency
    Date               
    2020-12-31     <NA>
    2020-01-20     1000
    """
    testing_df = dataframe_build(input_data)

    # then
    if dfbuild_type is DFBuildType.INDEX:
        assert isinstance(testing_df["Currency"][0], np.int64)
        assert testing_df["Currency"][1] is pd.NA

    elif dfbuild_type is DFBuildType.DATE_AS_INDEX:
        assert testing_df["Currency"][0] is pd.NA
        assert isinstance(testing_df["Currency"][1], np.int64)

    else:
        assert False, f"No assert for {dfbuild_type}"


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, EMPTY_STRING_LEAVE_AS_IT_IS_RDP_DATA),
        (ContentType.DATA_GRID_UDF, EMPTY_STRING_LEAVE_AS_IT_IS_UDF_DATA),
    ],
    ids=["Empty string leave as it is. RDP", "Empty string leave as it is. UDF"],
)
def test_empty_string_leave_as_it_is(content_type, input_data, dfbuild_type):
    # when
    dataframe_build = get_dfbuilder(content_type, dfbuild_type)
    """
    #### INDEX ####

      Instrument       Date Currency
    0     GOOG.O 2020-01-20         
    1     GOOG.O 2020-12-31         

    #### DATE_AS_INDEX ####

    GOOG.O     Currency
    Date               
    2020-12-31         
    2020-01-20         
       
    """
    testing_df = dataframe_build(input_data)

    # then
    assert testing_df["Currency"][0] == ""
    assert testing_df["Currency"][1] == ""


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, INSTRUMENT_AND_CURRENCY_ARE_EMPTY_STRINGS_RDP_DATA),
        (ContentType.DATA_GRID_UDF, INSTRUMENT_AND_CURRENCY_ARE_EMPTY_STRINGS_UDF_DATA),
    ],
    ids=[
        "Instrument and Currency are empty strings for dfbuild index. RDP",
        "Instrument and Currency are empty strings for dfbuild index. UDF",
    ],
)
def test_instrument_and_currency_are_empty_strings_for_dfbuild_index(
    content_type, input_data
):
    # when
    dataframe_build = get_dfbuilder(content_type, DFBuildType.INDEX)
    """
    #### INDEX ####
    
      Instrument       Date Currency
    0            2020-01-20     <NA>   
    1     GOOG.O 2020-12-31     1000
    """
    testing_df = dataframe_build(input_data)

    # then
    assert testing_df["Currency"][0] is pd.NA
    assert testing_df["Instrument"][0] == ""


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, INSTRUMENT_AND_CURRENCY_ARE_EMPTY_STRINGS_RDP_DATA),
        (ContentType.DATA_GRID_UDF, INSTRUMENT_AND_CURRENCY_ARE_EMPTY_STRINGS_UDF_DATA),
    ],
    ids=[
        "Instrument and Currency are empty strings for dfbuild date as index. RDP",
        "Instrument and Currency are empty strings for dfbuild date as index. UDF",
    ],
)
def test_instrument_and_currency_are_empty_strings_for_dfbuild_date_as_index(
    content_type, input_data
):
    expected_str = (
        "Currency          GOOG.O\n"
        "Date                    \n"
        "2020-12-31  <NA>    1000\n"
        "2020-01-20          <NA>"
    )

    # when
    dataframe_build = get_dfbuilder(content_type, DFBuildType.DATE_AS_INDEX)
    """
    #### DATE_AS_INDEX ####

    Currency          GOOG.O
    Date                    
    2020-12-31  <NA>    1000
    2020-01-20          <NA>
    """
    testing_df = dataframe_build(input_data)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_str, testing_str
    assert str(testing_df.dtypes["GOOG.O"]) == "Int64"
    assert str(testing_df.dtypes[""]) == "string"


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, TWO_UNIVERSES_ONE_FIELD_RDP_DATA),
        (ContentType.DATA_GRID_UDF, TWO_UNIVERSES_ONE_FIELD_UDF_DATA),
    ],
    ids=["Two universes and one field. RDP", "Two universes and one field. UDF"],
)
def test_two_universes_with_one_field(content_type, input_data):
    expected_str = (
        "BID         EUR=  LSEG.L\n"
        "Date                    \n"
        "2022-02-28  2000    <NA>\n"
        "2022-01-31  2000    <NA>\n"
        "2020-12-31  <NA>    1000\n"
        "2020-01-20  <NA>    1000"
    )
    # when
    dataframe_build = get_dfbuilder(content_type, DFBuildType.DATE_AS_INDEX)
    """
    #### DATE_AS_INDEX ####

    BID         EUR=  LSEG.L
    Date                    
    2022-02-28  2000    <NA>
    2022-01-31  2000    <NA>
    2020-12-31  <NA>    1000
    2020-01-20  <NA>    1000
    """
    testing_df = dataframe_build(input_data)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_str, testing_str
    assert str(testing_df.dtypes["EUR="]) == "Int64"
    assert str(testing_df.dtypes["LSEG.L"]) == "Int64"


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, TWO_SAME_DATES_DOES_NOT_MERGE_IN_ONE_RDP_DATA),
        (ContentType.DATA_GRID_UDF, TWO_SAME_DATES_DOES_NOT_MERGE_IN_ONE_UDF_DATA),
    ],
    ids=[
        "Two same values does not merge in one. RDP",
        "Two same values does not merge in one. UDF",
    ],
)
def test_two_same_dates_does_not_merge_in_one(content_type, input_data):
    expected_str = (
        "IBM        Currency  Revenue - Mean\n"
        "Date                               \n"
        "2021-08-30      USD     75127177530\n"
        "2021-07-29      USD     75129376130\n"
        "2021-05-21      USD     74329699000\n"
        "2021-05-21      USD     74329699000\n"
        "2021-04-25      USD     74397055190\n"
        "2021-03-22      USD     74221312380\n"
        "2021-02-11      USD     74195199870\n"
        "2021-01-27      USD     74195199870\n"
        "2020-12-16      USD     73950729070\n"
        "2020-10-23      USD     73965242400\n"
        "2020-10-23      USD     73965242400"
    )
    # when
    dataframe_build = get_dfbuilder(content_type, DFBuildType.DATE_AS_INDEX)
    """
    #### DATE_AS_INDEX ####

    IBM        Currency  Revenue - Mean
    Date                               
    2021-08-30      USD     75127177530
    2021-07-29      USD     75129376130
    2021-05-21      USD     74329699000
    2021-05-21      USD     74329699000
    2021-04-25      USD     74397055190
    2021-03-22      USD     74221312380
    2021-02-11      USD     74195199870
    2021-01-27      USD     74195199870
    2020-12-16      USD     73950729070
    2020-10-23      USD     73965242400
    2020-10-23      USD     73965242400
    """
    testing_df = dataframe_build(input_data)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_str, testing_str


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA),
        (ContentType.DATA_GRID_UDF, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA),
    ],
    ids=[
        "Merge values with same date in one row. RDP",
        "Merge values with same date in one row. UDF",
    ],
)
def test_merge_values_with_same_date_in_one_row(content_type, input_data):
    expected_str = (
        "                IBM                   VOD.L               \n"
        "           Currency Revenue - Mean Currency Revenue - Mean\n"
        "Date                                                      \n"
        "2021-08-30      USD    75127177530     <NA>           <NA>\n"
        "2021-08-25     <NA>           <NA>      EUR    45088444860\n"
        "2021-07-29      USD    75129376130      EUR    45109776750\n"
        "2021-06-29     <NA>           <NA>      EUR    44967283500\n"
        "2021-05-27     <NA>           <NA>      EUR    44918585000\n"
        "2021-05-21      USD    74329699000     <NA>           <NA>\n"
        "2021-05-21      USD    74329699000     <NA>           <NA>\n"
        "2021-04-28     <NA>           <NA>      EUR    43503348730\n"
        "2021-04-25      USD    74397055190     <NA>           <NA>\n"
        "2021-03-31     <NA>           <NA>      EUR    43445947190\n"
        "2021-03-22      USD    74221312380     <NA>           <NA>\n"
        "2021-02-24     <NA>           <NA>      EUR    43421860050\n"
        "2021-02-11      USD    74195199870     <NA>           <NA>\n"
        "2021-01-27      USD    74195199870     <NA>           <NA>\n"
        "2021-01-20     <NA>           <NA>      EUR    43333798760\n"
        "2020-12-16      USD    73950729070      EUR    43372965910\n"
        "2020-11-23     <NA>           <NA>      EUR    43442990380\n"
        "2020-10-23      USD    73965242400     <NA>           <NA>\n"
        "2020-10-23      USD    73965242400     <NA>           <NA>\n"
        "2020-10-21     <NA>           <NA>      EUR    43375027760"
    )
    # when
    dataframe_build = get_dfbuilder(content_type, DFBuildType.DATE_AS_INDEX)
    """
    #### DATE_AS_INDEX ####

                    IBM                   VOD.L               
               Currency Revenue - Mean Currency Revenue - Mean
    Date                                                      
    2021-08-30      USD    75127177530     <NA>           <NA>
    2021-08-25     <NA>           <NA>      EUR    45088444860
    2021-07-29      USD    75129376130      EUR    45109776750
    2021-06-29     <NA>           <NA>      EUR    44967283500
    2021-05-27     <NA>           <NA>      EUR    44918585000
    2021-05-21      USD    74329699000     <NA>           <NA>
    2021-05-21      USD    74329699000     <NA>           <NA>
    2021-04-28     <NA>           <NA>      EUR    43503348730
    2021-04-25      USD    74397055190     <NA>           <NA>
    2021-03-31     <NA>           <NA>      EUR    43445947190
    2021-03-22      USD    74221312380     <NA>           <NA>
    2021-02-24     <NA>           <NA>      EUR    43421860050
    2021-02-11      USD    74195199870     <NA>           <NA>
    2021-01-27      USD    74195199870     <NA>           <NA>
    2021-01-20     <NA>           <NA>      EUR    43333798760
    2020-12-16      USD    73950729070      EUR    43372965910
    2020-11-23     <NA>           <NA>      EUR    43442990380
    2020-10-23      USD    73965242400     <NA>           <NA>
    2020-10-23      USD    73965242400     <NA>           <NA>
    2020-10-21     <NA>           <NA>      EUR    43375027760
    """
    testing_df = dataframe_build(input_data)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_str, testing_str


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, DELETE_ROW_IF_DATE_IS_NONE_RDP_DATA),
        (ContentType.DATA_GRID_UDF, DELETE_ROW_IF_DATE_IS_NONE_UDF_DATA),
    ],
    ids=[
        "Delete row if date is none. RDP",
        "Delete row if date is none. UDF",
    ],
)
def test_delete_row_if_date_is_none(content_type, input_data):
    expected_str_rdp = (
        "                  GOOG.O                 INVAL                      FB.O             \n"
        "                 Revenue  Gross Profit Revenue Gross Profit      Revenue Gross Profit\n"
        "Date                                                                                 \n"
        "2021-12-31  257637000000  146698000000    <NA>         <NA>         <NA>         <NA>\n"
        "2020-12-31          <NA>          <NA>    <NA>         <NA>  85965000000  69273000000"
    )
    expected_str_udf = (
        "                  GOOG.O                               INVAL                                    FB.O                           \n"
        "                 Revenue  Gross Profit INVALID_FIELD Revenue Gross Profit INVALID_FIELD      Revenue Gross Profit INVALID_FIELD\n"
        "Date                                                                                                                           \n"
        "2021-12-31  257637000000  146698000000                  <NA>         <NA>          <NA>         <NA>         <NA>          <NA>\n"
        "2020-12-31          <NA>          <NA>          <NA>    <NA>         <NA>          <NA>  85965000000  69273000000              "
    )

    expected_str = (
        expected_str_udf
        if content_type is ContentType.DATA_GRID_UDF
        else expected_str_rdp
    )
    # when
    dataframe_build = get_dfbuilder(content_type, DFBuildType.DATE_AS_INDEX)
    """
    #### DATE_AS_INDEX ####
    # UDF #
                  GOOG.O                               INVAL                                    FB.O                           
                 Revenue  Gross Profit INVALID_FIELD Revenue Gross Profit INVALID_FIELD      Revenue Gross Profit INVALID_FIELD
Date                                                                                                                           
2021-12-31  257637000000  146698000000                  <NA>         <NA>          <NA>         <NA>         <NA>          <NA>
2020-12-31          <NA>          <NA>          <NA>    <NA>         <NA>          <NA>  85965000000  69273000000              

    # RDP #
                  GOOG.O                 INVAL                      FB.O             
                 Revenue  Gross Profit Revenue Gross Profit      Revenue Gross Profit
Date                                                                                 
2021-12-31  257637000000  146698000000    <NA>         <NA>         <NA>         <NA>
2020-12-31          <NA>          <NA>    <NA>         <NA>  85965000000  69273000000
    """
    testing_df = dataframe_build(input_data)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_str, f"\n{testing_df.to_string()}"


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA_1),
        (ContentType.DATA_GRID_UDF, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA_1),
    ],
    ids=[
        "Merge values with same date in one row. Combination 1. RDP",
        "Merge values with same date in one row. Combination 1. UDF",
    ],
)
def test_merge_values_with_same_date_in_one_row_combination_1(content_type, input_data):
    expected_str = (
        "                IBM                   VOD.L                   NKE.N               \n"
        "           Currency Revenue - Mean Currency Revenue - Mean Currency Revenue - Mean\n"
        "Date                                                                              \n"
        "2020-10-23      USD    73965242400      EUR    43375027760     <NA>           <NA>\n"
        "2020-05-07     <NA>           <NA>     <NA>           <NA>      GBP    43375027760"
    )

    # when
    dataframe_build = get_dfbuilder(content_type, DFBuildType.DATE_AS_INDEX)
    """
    #### DATE_AS_INDEX ####
                IBM                   VOD.L                   NKE.N               
           Currency Revenue - Mean Currency Revenue - Mean Currency Revenue - Mean
Date                                                                              
2020-10-23      USD    73965242400      EUR    43375027760     <NA>           <NA>
2020-05-07     <NA>           <NA>     <NA>           <NA>      GBP    43375027760
    """
    testing_df = dataframe_build(input_data)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_str, f"\n{testing_df.to_string()}"


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA_2),
        (ContentType.DATA_GRID_UDF, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA_2),
    ],
    ids=[
        "Merge values with same date in one row. Combination 2. RDP",
        "Merge values with same date in one row. Combination 2. UDF",
    ],
)
def test_merge_values_with_same_date_in_one_row_combination_2(content_type, input_data):
    expected_str = (
        "                IBM                   VOD.L                   NKE.N               \n"
        "           Currency Revenue - Mean Currency Revenue - Mean Currency Revenue - Mean\n"
        "Date                                                                              \n"
        "2020-10-23      USD    73965242400      EUR    43375027760      GBP    43375027760"
    )

    # when
    dataframe_build = get_dfbuilder(content_type, DFBuildType.DATE_AS_INDEX)
    """
    #### DATE_AS_INDEX ####
                IBM                   VOD.L                   NKE.N               
           Currency Revenue - Mean Currency Revenue - Mean Currency Revenue - Mean
Date                                                                              
2020-10-23      USD    73965242400      EUR    43375027760      GBP    43375027760
    """
    testing_df = dataframe_build(input_data)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_str, f"\n{testing_df.to_string()}"


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA_3),
        (ContentType.DATA_GRID_UDF, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA_3),
    ],
    ids=[
        "Merge values with same date in one row. Combination 3. RDP",
        "Merge values with same date in one row. Combination 3. UDF",
    ],
)
def test_merge_values_with_same_date_in_one_row_combination_3(content_type, input_data):
    expected_str = (
        "                IBM                   VOD.L                   NKE.N               \n"
        "           Currency Revenue - Mean Currency Revenue - Mean Currency Revenue - Mean\n"
        "Date                                                                              \n"
        "2020-10-23     <NA>           <NA>      EUR    43375027760     <NA>           <NA>\n"
        "2020-09-18      USD    73965242400     <NA>           <NA>     <NA>           <NA>\n"
        "2020-05-07     <NA>           <NA>     <NA>           <NA>      GBP    43375027760"
    )

    # when
    dataframe_build = get_dfbuilder(content_type, DFBuildType.DATE_AS_INDEX)
    """
    #### DATE_AS_INDEX ####
                IBM                   VOD.L                   NKE.N               
           Currency Revenue - Mean Currency Revenue - Mean Currency Revenue - Mean
Date                                                                              
2020-10-23     <NA>           <NA>      EUR    43375027760     <NA>           <NA>
2020-09-18      USD    73965242400     <NA>           <NA>     <NA>           <NA>
2020-05-07     <NA>           <NA>     <NA>           <NA>      GBP    43375027760
    """
    testing_df = dataframe_build(input_data)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_str, f"\n{testing_df.to_string()}"


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA_4),
        (ContentType.DATA_GRID_UDF, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA_4),
    ],
    ids=[
        "Merge values with same date in one row. Combination 4. RDP",
        "Merge values with same date in one row. Combination 4. UDF",
    ],
)
def test_merge_values_with_same_date_in_one_row_combination_4(content_type, input_data):
    expected_str = (
        "                IBM                   VOD.L                   NKE.N               \n"
        "           Currency Revenue - Mean Currency Revenue - Mean Currency Revenue - Mean\n"
        "Date                                                                              \n"
        "2020-10-23     <NA>           <NA>      EUR    43375027760      GBP    43375027760\n"
        "2020-05-07      USD    73965242400     <NA>           <NA>     <NA>           <NA>"
    )

    # when
    dataframe_build = get_dfbuilder(content_type, DFBuildType.DATE_AS_INDEX)
    """
    #### DATE_AS_INDEX ####
                IBM                   VOD.L                   NKE.N               
           Currency Revenue - Mean Currency Revenue - Mean Currency Revenue - Mean
Date                                                                              
2020-10-23     <NA>           <NA>      EUR    43375027760      GBP    43375027760
2020-05-07      USD    73965242400     <NA>           <NA>     <NA>           <NA>
    """
    testing_df = dataframe_build(input_data)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_str, f"\n{testing_df.to_string()}"


@pytest.mark.parametrize(
    "content_type, input_data",
    [
        (ContentType.DATA_GRID_RDP, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_RDP_DATA_5),
        (ContentType.DATA_GRID_UDF, MERGE_VALUES_WITH_SAME_DATE_IN_ONE_ROW_UDF_DATA_5),
    ],
    ids=[
        "Merge values with same date in one row. Combination 5. RDP",
        "Merge values with same date in one row. Combination 5. UDF",
    ],
)
def test_merge_values_with_same_date_in_one_row_combination_5(content_type, input_data):
    expected_str = (
        "                IBM                   VOD.L                   NKE.N               \n"
        "           Currency Revenue - Mean Currency Revenue - Mean Currency Revenue - Mean\n"
        "Date                                                                              \n"
        "2020-10-23      USD    73965242400     <NA>           <NA>      GBP    43375027760\n"
        "2020-05-07     <NA>           <NA>      EUR    43375027760     <NA>           <NA>"
    )

    # when
    dataframe_build = get_dfbuilder(content_type, DFBuildType.DATE_AS_INDEX)
    """
    #### DATE_AS_INDEX ####
                IBM                   VOD.L                   NKE.N               
           Currency Revenue - Mean Currency Revenue - Mean Currency Revenue - Mean
Date                                                                              
2020-10-23      USD    73965242400     <NA>           <NA>      GBP    43375027760
2020-05-07     <NA>           <NA>      EUR    43375027760     <NA>           <NA>
    """
    testing_df = dataframe_build(input_data)

    # then
    testing_str = testing_df.to_string()
    assert testing_str == expected_str, f"\n{testing_df.to_string()}"



# coding: utf8

from typing import Optional

from ..._object_definition import ObjectDefinition


class PricingParameters(ObjectDefinition):
    """
    API endpoint for Financial Contract analytics,
    that returns calculations relevant to each contract type.

    Parameters
    ----------
    cash_amount_in_deal_ccy : float, optional
        cash_amount_in_deal_ccy to override and that will be used as pricing analysis
        input to compute the cds other outputs. Optional. No override is applied by
        default. Note that only one pricing analysis input should be defined.
    clean_price_percent : float, optional
        clean_price_percent to override and that will be used as pricing analysis input
        to compute the cds other outputs. Optional. No override is applied by default.
        Note that only one pricing analysis input should be defined.
    conventional_spread_bp : float, optional
        conventional_spread_bp to override and that will be used as pricing analysis
        input to compute the cds other outputs. Optional. No override is applied by
        default. Note that only one pricing analysis input should be defined.
    market_data_date : str, optional
        The market data date for pricing. Optional. By default, the market_data_date
        date is the valuation_date or Today
    report_ccy : str, optional
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). it is set for the fields ending with 'xxxinreportccy'. optional. the
        default value is the notional currency.
    upfront_amount_in_deal_ccy : float, optional
        upfront_amount_in_deal_ccy to override and that will be used as pricing analysis
        input to compute the cds other outputs. Optional. No override is applied by
        default. Note that only one pricing analysis input should be defined.
    upfront_percent : float, optional
        upfront_percent to override and that will be used as pricing analysis input to
        compute the cds other outputs. Optional. No override is applied by default. Note
        that only one pricing analysis input should be defined.
    valuation_date : str, optional
        The valuation date for pricing.  Optional. If not set the valuation date is
        equal to market_data_date or Today. For assets that contains a
        settlementConvention, the default valuation date  is equal to the settlementdate
        of the Asset that is usually the TradeDate+SettlementConvention.
    """

    def __init__(
        self,
        cash_amount_in_deal_ccy: Optional[float] = None,
        clean_price_percent: Optional[float] = None,
        conventional_spread_bp: Optional[float] = None,
        market_data_date: Optional[str] = None,
        report_ccy: Optional[str] = None,
        upfront_amount_in_deal_ccy: Optional[float] = None,
        upfront_percent: Optional[float] = None,
        valuation_date: Optional[str] = None,
    ) -> None:
        super().__init__()
        self.cash_amount_in_deal_ccy = cash_amount_in_deal_ccy
        self.clean_price_percent = clean_price_percent
        self.conventional_spread_bp = conventional_spread_bp
        self.market_data_date = market_data_date
        self.report_ccy = report_ccy
        self.upfront_amount_in_deal_ccy = upfront_amount_in_deal_ccy
        self.upfront_percent = upfront_percent
        self.valuation_date = valuation_date

    @property
    def cash_amount_in_deal_ccy(self):
        """
        cash_amount_in_deal_ccy to override and that will be used as pricing analysis
        input to compute the cds other outputs. Optional. No override is applied by
        default. Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("cashAmountInDealCcy")

    @cash_amount_in_deal_ccy.setter
    def cash_amount_in_deal_ccy(self, value):
        self._set_parameter("cashAmountInDealCcy", value)

    @property
    def clean_price_percent(self):
        """
        clean_price_percent to override and that will be used as pricing analysis input
        to compute the cds other outputs. Optional. No override is applied by default.
        Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("cleanPricePercent")

    @clean_price_percent.setter
    def clean_price_percent(self, value):
        self._set_parameter("cleanPricePercent", value)

    @property
    def conventional_spread_bp(self):
        """
        conventional_spread_bp to override and that will be used as pricing analysis
        input to compute the cds other outputs. Optional. No override is applied by
        default. Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("conventionalSpreadBp")

    @conventional_spread_bp.setter
    def conventional_spread_bp(self, value):
        self._set_parameter("conventionalSpreadBp", value)

    @property
    def market_data_date(self):
        """
        The market data date for pricing. Optional. By default, the market_data_date
        date is the valuation_date or Today
        :return: str
        """
        return self._get_parameter("marketDataDate")

    @market_data_date.setter
    def market_data_date(self, value):
        self._set_parameter("marketDataDate", value)

    @property
    def report_ccy(self):
        """
        The reporting currency code, expressed in iso 4217 alphabetical format (e.g.,
        'usd'). It is set for the fields ending with 'xxxinreportccy'. Optional. The
        default value is the notional currency.
        :return: str
        """
        return self._get_parameter("reportCcy")

    @report_ccy.setter
    def report_ccy(self, value):
        self._set_parameter("reportCcy", value)

    @property
    def upfront_amount_in_deal_ccy(self):
        """
        upfront_amount_in_deal_ccy to override and that will be used as pricing analysis
        input to compute the cds other outputs. Optional. No override is applied by
        default. Note that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("upfrontAmountInDealCcy")

    @upfront_amount_in_deal_ccy.setter
    def upfront_amount_in_deal_ccy(self, value):
        self._set_parameter("upfrontAmountInDealCcy", value)

    @property
    def upfront_percent(self):
        """
        upfront_percent to override and that will be used as pricing analysis input to
        compute the cds other outputs. Optional. No override is applied by default. Note
        that only one pricing analysis input should be defined.
        :return: float
        """
        return self._get_parameter("upfrontPercent")

    @upfront_percent.setter
    def upfront_percent(self, value):
        self._set_parameter("upfrontPercent", value)

    @property
    def valuation_date(self):
        """
        The valuation date for pricing.  Optional. If not set the valuation date is
        equal to market_data_date or Today. For assets that contains a
        settlementConvention, the default valuation date  is equal to the settlementdate
        of the Asset that is usually the TradeDate+SettlementConvention.
        :return: str
        """
        return self._get_parameter("valuationDate")

    @valuation_date.setter
    def valuation_date(self, value):
        self._set_parameter("valuationDate", value)



from collections import namedtuple
from dataclasses import dataclass, field
from datetime import timezone, datetime as dt
from enum import Enum
from typing import Dict, Tuple, Set, Type, TYPE_CHECKING

if TYPE_CHECKING:
    from ._filter_types import FilterType
    from ._abstract_logger import AbstractUsageLogger


def _get_utc_time() -> float:
    return dt.now(timezone.utc).timestamp()


@dataclass(frozen=True)
class RecordData:
    """
    This class is used to store the data for a single log record.
    """

    args: Tuple = field(default_factory=tuple)
    kwargs: Dict = field(default_factory=dict)
    result: object = field(default=None)


@dataclass(frozen=True)
class LoggerConfig:
    """
    This class is used to store the configuration for a single user logger.
    """

    logger_type: Type["AbstractUsageLogger"]
    args: Tuple = field(default_factory=tuple)
    kwargs: Dict = field(default_factory=dict)
    filters: Set["FilterType"] = field(default_factory=set)


@dataclass(frozen=True)
class LogRecord:
    """
    This class is used to store record, with filter used by logger and name of
    the record.
    """

    name: str
    func_path: str
    data: RecordData
    filter: Set["FilterType"] = field(default_factory=set)
    timestamp: float = field(default_factory=_get_utc_time)


class ModuleName(str, Enum):
    DELIVERY = "Delivery"
    CONTENT = "Content"
    ACCESS = "Access"

    def __str__(self):
        return self.value


StreamUsageKey = namedtuple("StreamUsageKey", ["stream_id", "service", "name"])



import datetime

import pytest
from pandas.core.dtypes.common import is_datetime64_any_dtype

from refinitiv.data._core.session import set_default
from refinitiv.data._tools import get_params, ParamItem
from refinitiv.data.content import custom_instruments as ci
from refinitiv.data._errors import RDError
from refinitiv.data.content import custom_instruments
from refinitiv.data._content_type import ContentType
from refinitiv.data.content.custom_instruments._custom_instruments_data_provider import (
    get_content_type_by_interval,
    get_user_id,
    CustomInstsRequestFactory,
    CustomInstsSearchRequestFactory,
    custom_instruments_intervals_build_df,
    convert_to_holidays,
)
from refinitiv.data.delivery._data._endpoint_data import RequestMethod
from refinitiv.data.delivery._data._request import Request
from tests.unit.conftest import StubSession, StubResponse, args
from tests.unit.conftest import StubSession, StubResponse
from tests.unit.content.custom_instruments.conftest import (
    volume_based_udc,
    volume_based_dict_udc,
    day_based_udc,
    day_based_dict_udc,
    manual_udc,
    manual_dict_udc,
    basket_dict,
    ci_udc_response,
    ci_basket_response,
    ci_create_formula_response,
    basket_obj,
)
from tests.unit.content.custom_instruments.test__stream_facade import get_uuid_url


@pytest.mark.parametrize(
    ("input_params_details", "input_data", "test_result"),
    [
        (
            [ParamItem("key", "KEY")],
            {"key": "value"},
            [
                ("KEY", "value"),
            ],
        ),
        (
            [ParamItem("key", "KEY")],
            {"key": "value", "dasd": "1"},
            [
                ("KEY", "value"),
            ],
        ),
        (
            [ParamItem("key", "KEY", lambda a, **_: a.upper())],
            {"key": "value"},
            [
                ("KEY", "VALUE"),
            ],
        ),
    ],
)
def test_get_params(input_params_details, input_data, test_result):
    # given
    # when
    result = get_params(input_params_details, **input_data)
    # then
    assert result == test_result


@pytest.mark.parametrize(
    ("input_interval", "test_result"),
    [
        ("PT1M", ContentType.CUSTOM_INSTRUMENTS_INTRADAY_SUMMARIES),
        ("P1D", ContentType.CUSTOM_INSTRUMENTS_INTERDAY_SUMMARIES),
    ],
)
def test_get_content_type_by_interval(input_interval, test_result):
    # given
    # when
    result = get_content_type_by_interval(input_interval)
    # then
    assert result == test_result


@pytest.mark.parametrize(
    ("content_data", "expected_uuid"),
    [
        args(
            content_data=[
                {
                    "state": {
                        "code": 400,
                        "status": "Bad Request",
                        "message": "Validation Error",
                    },
                    "data": [
                        {
                            "key": "symbol",
                            "reason": f".UUID suffix UUID-0000 not matched with userID GE-1234",
                        }
                    ],
                }
            ],
            expected_uuid="GE-1234",
        ),
        args(
            content_data=[
                {
                    "state": {
                        "code": 400,
                        "status": "Bad Request",
                        "message": "Validation Error",
                    },
                    "data": [
                        {
                            "key": "symbol",
                            "reason": ".UUID suffix UUID-0000 not matched with userID SL1-1RZ2M3V",
                        }
                    ],
                }
            ],
            expected_uuid="SL1-1RZ2M3V",
        ),
    ],
)
def test_get_user_id(content_data, expected_uuid):
    # given
    def _http_request(request: Request):
        if request.url == get_uuid_url:
            return StubResponse(
                content_data=content_data,
                status_code=content_data[0].get("state", {}).get("code", -1),
            )

    session = StubSession(is_open=True)
    session.http_request = _http_request

    # when
    testing_uuid = get_user_id(session=session)

    # then
    assert testing_uuid == expected_uuid


@pytest.mark.parametrize(
    ("content_data", "test_result"),
    [
        ([{}], ""),
    ],
)
def test_get_uuid_error(content_data, test_result):
    def _http_request(request: Request):
        if request.url == get_uuid_url:
            return StubResponse(
                content_data=content_data,
                status_code=content_data[0].get("state", {}).get("code", -1),
            )

    session = StubSession(is_open=True)
    session.http_request = _http_request

    # then
    with pytest.raises(RDError):
        # when
        get_user_id(session=session)


def test_get_user_id_when_not_expected_error():
    session = StubSession(
        is_open=True, response=StubResponse(content_data="", status_code=401)
    )
    with pytest.raises(RDError):
        get_user_id(session)


@pytest.mark.parametrize(
    ("input_kwargs", "request_method", "test_result"),
    [
        (
            {"symbol": "MyInstrument", "description": "Description"},
            RequestMethod.GET,
            {},
        ),
        (
            {
                "symbol": "MyInstrument",
                "currency": "GBP",
                "description": "Description",
                "exchange_name": "ADSE",
                "formula": "VOD.L*3",
                "holidays": [{"reason": "New Year", "date": "1.01.2022"}],
                "time_zone": "LON",
                "instrument_name": "InstrName",
            },
            RequestMethod.POST,
            {
                "symbol": "S)MyInstrument.",
                "currency": "GBP",
                "description": "Description",
                "exchangeName": "ADSE",
                "formula": "VOD.L*3",
                "holidays": [{"reason": "New Year", "date": "1.01.2022"}],
                "timeZone": "LON",
                "instrumentName": "InstrName",
            },
        ),
    ],
)
def test_custom_insts_request_factory_get_body_parameters(
    input_kwargs, request_method, test_result
):
    # given
    request_factory = CustomInstsRequestFactory()
    request_factory.get_request_method = lambda *args, **kwargs: request_method
    session = StubSession(is_open=True)
    set_default(session)

    # when
    result = request_factory.get_body_parameters(session, **input_kwargs)

    # then
    assert result == test_result


def test_search_request_factory_get_query_parameters():
    # given
    params = {"access": "owner"}
    test_result = [("access", "owner")]

    # when
    request_factory = CustomInstsSearchRequestFactory()
    result = request_factory.get_query_parameters(**params)

    # then
    assert result == test_result


def test_custom_instruments_intervals_build_df_datetime_type():
    # given
    testing_column_name = "DATE"
    input_content_data = {
        "data": [["2012-11-01T04:16:13"]],
        "headers": [{"name": testing_column_name}],
    }

    # when
    result_df = custom_instruments_intervals_build_df(input_content_data)

    # then
    assert is_datetime64_any_dtype(result_df[testing_column_name])


summaries_response = [
    {
        "data": [
            ["2022-09-27", 2.8776],
            ["2022-09-26", 2.8818],
            ["2022-09-23", 2.907],
            ["2022-09-22", 2.9508],
            ["2022-09-21", 2.9511],
            ["2022-09-20", 2.991],
            ["2022-09-19", 3.0066],
            ["2022-09-16", 3.0045],
            ["2022-09-15", 2.9997],
            ["2022-09-14", 2.9931],
            ["2022-09-13", 2.991],
            ["2022-09-12", 3.0357],
            ["2022-09-09", 3.0117],
            ["2022-09-08", 2.9982],
            ["2022-09-07", 2.9997],
            ["2022-09-06", 2.9706],
            ["2022-09-05", 2.9778],
            ["2022-09-02", 2.9853],
            ["2022-09-01", 2.9832],
            ["2022-08-31", 3.0171],
        ],
        "defaultPricingField": "TRDPRC_1",
        "headers": [
            {"name": "DATE", "type": "string"},
            {"decimalChar": ".", "name": "TRDPRC_1", "type": "number"},
        ],
        "interval": "P1D",
        "summaryTimestampLabel": "endPeriod",
        "universe": {"ric": "S)TR12.GESG1-106493"},
    }
]
events_reponse = [
    {
        "data": [
            ["2022-09-28T06:33:57.622Z", 2.8692],
            ["2022-09-28T06:33:57.062Z", 2.8695],
            ["2022-09-28T06:33:56.052Z", 2.8695],
            ["2022-09-28T06:33:55.595Z", 2.8695],
            ["2022-09-28T06:33:55.255Z", 2.8692],
            ["2022-09-28T06:33:54.585Z", 2.8695],
            ["2022-09-28T06:33:54.167Z", 2.8698],
            ["2022-09-28T06:33:53.598Z", 2.8692],
            ["2022-09-28T06:33:53.086Z", 2.8698],
            ["2022-09-28T06:33:52.586Z", 2.8695],
            ["2022-09-28T06:33:51.583Z", 2.8692],
            ["2022-09-28T06:33:50.054Z", 2.8692],
            ["2022-09-28T06:33:48.432Z", 2.8692],
            ["2022-09-28T06:33:47.971Z", 2.8692],
            ["2022-09-28T06:33:47.751Z", 2.8698],
            ["2022-09-28T06:33:47.555Z", 2.8692],
            ["2022-09-28T06:33:46.782Z", 2.8692],
            ["2022-09-28T06:33:46.544Z", 2.8692],
            ["2022-09-28T06:33:45.117Z", 2.8698],
            ["2022-09-28T06:33:44.554Z", 2.8692],
        ],
        "defaultPricingField": "TRDPRC_1",
        "headers": [
            {"name": "DATE_TIME", "type": "string"},
            {"decimalChar": ".", "name": "TRDPRC_1", "type": "number"},
        ],
        "interval": None,
        "summaryTimestampLabel": None,
        "universe": {"ric": "S)TR12.GESG1-106493"},
    }
]


def test_fields_events():
    # given
    session = StubSession(is_open=True, response=StubResponse(events_reponse))

    # when
    response = custom_instruments.events.Definition(
        universe=["S)TR12"], fields=["TRDPRC_1"]
    ).get_data(session=session)
    df = response.data.df

    # then
    assert "TRDPRC_1" in df


def test_fields_events_empty_df():
    # given
    session = StubSession(is_open=True, response=StubResponse(events_reponse))

    # when
    response = custom_instruments.events.Definition(
        universe=["S)TR12"], fields=["BID"]
    ).get_data(session=session)
    df = response.data.df

    # then
    assert df.empty


def test_fields_summaries():
    # given
    session = StubSession(is_open=True, response=StubResponse(summaries_response))

    # when
    response = custom_instruments.summaries.Definition(
        universe=["S)TR12"], fields=["TRDPRC_1"]
    ).get_data(session=session)
    df = response.data.df

    # then
    assert "TRDPRC_1" in df


def test_fields_summaries_empty_df():
    # given
    session = StubSession(is_open=True, response=StubResponse(summaries_response))

    # when
    response = custom_instruments.summaries.Definition(
        universe=["S)TR12"], fields=["BID"]
    ).get_data(session=session)
    df = response.data.df

    # then
    assert df.empty


@pytest.mark.parametrize(
    ("ci_object", "ci_dict"),
    [
        (volume_based_udc, volume_based_dict_udc),
        (day_based_udc, day_based_dict_udc),
        (manual_udc, manual_dict_udc),
        (basket_obj, basket_dict),
    ],
    ids=["volumeBased_udc", "dayBased_udc", "manual_udc", "basket"],
)
def test_custom_instrument_serialize_into_json(ci_object, ci_dict):
    # given
    ci_cls = ci_object

    # when
    converted_ci_to_dict = ci_cls._to_dict()

    # then
    assert ci_dict == converted_ci_to_dict


@pytest.mark.parametrize(
    ("ci_object", "ci_dict"),
    [
        (volume_based_udc, volume_based_dict_udc),
        (day_based_udc, day_based_dict_udc),
        (basket_obj, basket_dict),
    ],
    ids=["volumeBased_udc", "dayBased_udc", "basket"],
)
def test_custom_instrument_deserialize_into_python_obj(ci_object, ci_dict):
    # given
    ci_cls = ci_object

    # when
    converted_ci_from_dict = ci_cls._from_dict(ci_dict)

    # then
    assert ci_cls == converted_ci_from_dict


def test_create_udc_custom_instrument():
    # given
    session = StubSession(is_open=True, response=StubResponse(ci_udc_response))

    # when
    instrument = ci.manage.create_udc(
        symbol="My_UDC_instrument",
        instrument_name="9789",
        exchange_name="5050",
        currency="GBP",
        time_zone="LON",
        description="fintech trading tool",
        holidays=[
            ci.manage.Holiday(date="1991-08-23", name="Independence Day of Ukraine"),
            {"date": "2022-12-18", "reason": "Hanukkah"},
        ],
        udc=volume_based_udc,
        session=session,
    )

    # then
    assert instrument.symbol == ci_udc_response.get("symbol")
    assert instrument.udc._to_dict() == ci_udc_response.get("udc")
    assert instrument.exchange_name == ci_udc_response.get("exchangeName")
    assert instrument.holidays == ci_udc_response.get("holidays")


def test_create_basket_custom_instrument():
    # given
    session = StubSession(is_open=True, response=StubResponse(ci_basket_response))

    # when
    instrument = ci.manage.create_basket(
        symbol="My_Basket_instrument",
        instrument_name="9789",
        exchange_name="5050",
        currency="GBP",
        time_zone="LON",
        description="fintech trading tool",
        holidays=[
            ci.manage.Holiday(date="1991-08-23", name="Independence Day of Ukraine"),
            {"date": "2022-12-18", "reason": "Hanukkah"},
        ],
        basket=basket_obj,
        session=session,
    )

    # then
    assert instrument.symbol == ci_basket_response.get("symbol")
    assert instrument.basket._to_dict() == ci_basket_response.get("basket")
    assert instrument.instrument_name == ci_basket_response.get("instrumentName")
    assert instrument.time_zone == ci_basket_response.get("timeZone")


def test_formula_custom_instrument_getting_the_same_obj():
    # given
    session = StubSession(
        is_open=True, response=StubResponse(ci_create_formula_response)
    )

    # when
    instrument = ci.manage.create_formula(
        symbol="My_formula_instrument",
        instrument_name="9789",
        exchange_name="5050",
        currency="GBP",
        time_zone="LON",
        description="fintech trading tool",
        holidays=[
            ci.manage.Holiday(date="1991-08-23", name="Independence Day of Ukraine"),
            {"date": "2022-12-18", "reason": "Hanukkah"},
        ],
        formula="GBP=*3",
        session=session,
    )

    requested_instrument = ci.manage.get(
        universe="My_formula_instrument", session=session
    )

    # then
    assert instrument.symbol == requested_instrument.symbol
    assert instrument.formula == requested_instrument.formula
    assert instrument.currency == requested_instrument.currency
    assert instrument.holidays == requested_instrument.holidays


@pytest.mark.parametrize(
    ("changed_params"),
    [
        ({"formula": "EUR=*3", "currency": "EUR", "exchange_name": "9898"}),
    ],
)
def test_update_formula_custom_instrument(changed_params):
    # given
    session = StubSession(
        is_open=True, response=StubResponse(ci_create_formula_response)
    )

    # when
    instrument = ci.manage.get(universe="My_formula_instrument", session=session)
    instrument.formula = "EUR=*3"
    instrument.currency = "EUR"
    instrument.exchange_name = "9898"
    instrument.save()

    requested_instrument = ci.manage.get(
        universe="My_formula_instrument", session=session
    )

    # then
    assert instrument.formula == changed_params["formula"]
    assert instrument.currency == changed_params["currency"]
    assert instrument.exchange_name == changed_params["exchange_name"]
    assert instrument.formula == requested_instrument.formula
    assert instrument.currency == requested_instrument.currency
    assert instrument.exchange_name == requested_instrument.exchange_name


def test_param_item_arg_name_query_param_name():
    # given
    param_item = ParamItem(arg_name="name", query_param_name="NAME")

    # then
    assert param_item.arg_name == "name"
    assert param_item.query_param_name == "NAME"


def test_param_item_arg_name():
    # given
    param_item = ParamItem(arg_name="name")

    # then
    assert param_item.arg_name == "name"
    assert param_item.query_param_name == "name"


def test_convert_holiday_obj_to_dict():
    # given
    holidays = [
        ci.manage.Holiday(date="1991-10-24", name="Labour Day"),
        ci.manage.Holiday(
            date=datetime.date(2021, 8, 24), name="Independence Day of Ukraine"
        ),
        ci.manage.Holiday(date=datetime.timedelta(days=-30), name="Alaska Day"),
        {"date": "2022-04-23", "reason": "Shakespeare Day"},
    ]

    # when
    converted_holidays = convert_to_holidays(holidays)

    # then
    assert len(converted_holidays) == 4
    for holiday in converted_holidays:
        assert isinstance(holiday, dict)
        assert "date" in holiday and "reason" in holiday
        assert isinstance(holiday.get("date"), str)
        assert isinstance(holiday.get("reason"), str)



from refinitiv.data.content.custom_instruments import summaries


def get_summaries_definition_with_one_universe(create_instrument, interval):
    instrument_01 = create_instrument()
    response = summaries.Definition(universe=instrument_01, interval=interval)
    return response, instrument_01


def get_summaries_definition_with_list_universes(create_instrument, interval):
    instrument_01 = create_instrument(
        type_="basket",
        basket={
            "constituents": [
                {"ric": "IBM.N", "weight": 20},
                {"ric": "EPAM.N", "weight": 80},
            ],
            "normalizeByWeight": True,
        },
    )
    instrument_02 = create_instrument()
    response = summaries.Definition(
        universe=[instrument_01, instrument_02], interval=interval
    )
    return response, [instrument_01, instrument_02]



from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    CrossCurrencyCurveParameters,
)
from refinitiv.data.content.ipa._enums._extrapolation_mode import ExtrapolationMode
from refinitiv.data.content.ipa._curves._cross_currency_curves._enums import (
    InterpolationMode,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._definitions import (
    TurnAdjustment,
)


def test_cross_currency_curve_parameters():
    # given
    expected_dict = {
        "extrapolationMode": "Constant",
        "interpolationMode": "CubicSpline",
        "turnAdjustment": {},
    }

    # when
    testing_obj = CrossCurrencyCurveParameters(
        extrapolation_mode=ExtrapolationMode.CONSTANT,
        interpolation_mode=InterpolationMode.CUBIC_SPLINE,
        turn_adjustment=TurnAdjustment(),
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.extrapolation_mode == ExtrapolationMode.CONSTANT
    assert testing_obj.interpolation_mode == InterpolationMode.CUBIC_SPLINE
    assert testing_obj.turn_adjustment == TurnAdjustment()



from typing import TYPE_CHECKING

from pandas import DataFrame

from .._tools import (
    ParamItem,
    SerParamItem,
    qpl_datetime_adapter,
    ValueParamItem,
    is_date_true,
)
from ..content._content_data import Data as BaseData
from ..content._content_response_factory import ContentResponseFactory
from ..delivery._data import RequestMethod
from ..delivery._data._data_provider import DataProvider
from ..delivery._data._request_factory import RequestFactory
from ..delivery._data._validators import ValidatorContainer, ContentValidator

if TYPE_CHECKING:
    from ..delivery._data._parsed_data import ParsedData

fx_swp_to_swp_body_params_config = [
    ParamItem("fx_cross_code", "fxCrossCode"),
    ValueParamItem(
        "market_data_date_time",
        "marketDataDate",
        qpl_datetime_adapter.get_str,
        is_true=is_date_true,
    ),
    ParamItem("tenors"),
    ParamItem("fields"),
    SerParamItem("spot_ccy_1", "spotCcy1"),
    SerParamItem("spot_ccy_2", "spotCcy2"),
    SerParamItem("swap_points_ccy_1", "swapPointsCcy1"),
    SerParamItem("swap_points_ccy_2", "swapPointsCcy2"),
]


class FxSwpToSwpRequestFactory(RequestFactory):
    def update_url(self, url_root, url, path_parameters, query_parameters) -> str:
        return url

    def get_request_method(self, *, method=None, **kwargs) -> RequestMethod:
        return RequestMethod.POST

    @property
    def body_params_config(self):
        return fx_swp_to_swp_body_params_config

    def get_header_parameters(self, *args, **kwargs) -> dict:
        return {
            "x-tr-applicationid": "QPSInternal",
            "x-tr-clientapplicationid": "postman",
            "x-tr-clientid": "RDLib_python",
            "x-tr-loglevel": "Debug",
            "x-tr-scope": "RDLib-scope",
            "x-tr-uuid": "PADACT-002",
        }


class Data(BaseData):
    @property
    def df(self):
        if self._dataframe is None:
            swap_points = self.raw["swapPoints"]
            self._dataframe = DataFrame(swap_points)

        return self._dataframe


class QPLContentValidator(ContentValidator):
    @classmethod
    def content_data_has_no_error(cls, data: "ParsedData") -> bool:
        content_data = data.content_data
        error = content_data.get("error")
        if error:
            data.error_codes = error.get("code")
            data.error_messages = error.get("message")
            return False

        return True

    def __init__(self) -> None:
        super().__init__()
        self.validators.append(self.content_data_has_no_error)


fx_swp_to_swp_data_provider = DataProvider(
    request=FxSwpToSwpRequestFactory(),
    response=ContentResponseFactory(data_class=Data),
    validator=ValidatorContainer(content_validator=QPLContentValidator()),
)



import pytest

from refinitiv.data.content import pricing, trade_data_service
from refinitiv.data.content.ipa import financial_contracts
from refinitiv.data.content.pricing import chain
from refinitiv.data.delivery import rdp_stream, omm_stream
from refinitiv.data import OpenState
from tests.unit.conftest import StubSession


def test_session_open_state():
    import refinitiv.data as rd

    try:
        rd.OpenState
    except Exception as e:
        assert False, str(e)


@pytest.mark.parametrize(
    "definition",
    [
        chain.Definition("0#.FTSE"),
        pricing.Definition("EUR="),
        rdp_stream.Definition(
            service=None, universe=[], view=None, parameters={}, api="api"
        ),
        omm_stream.Definition("EUR"),
        trade_data_service.Definition(),
    ],
    ids=[
        "chain",
        "pricing",
        "rdp_stream",
        "omm_stream",
        "trade_data_service",
    ],
)
def test_stream_pending_state(definition):
    def do_open(*args, **kwargs):
        # then
        assert stream.open_state is OpenState.Pending

    # given
    session = StubSession(is_open=True)
    stream = definition.get_stream(session=session)
    stream._stream._do_open = do_open

    # when
    stream.open()


@pytest.mark.parametrize(
    "definition",
    [
        financial_contracts.bond.Definition("[]"),
        financial_contracts.cap_floor.Definition(),
        financial_contracts.cds.Definition(),
        financial_contracts.cross.Definition(),
        financial_contracts.option.Definition(),
        financial_contracts.repo.Definition(),
        financial_contracts.swap.Definition(),
        financial_contracts.swaption.Definition(),
        financial_contracts.term_deposit.Definition(),
    ],
    ids=[
        "financial_contracts.bond",
        "financial_contracts.cap_floor",
        "financial_contracts.cds",
        "financial_contracts.cross",
        "financial_contracts.option",
        "financial_contracts.repo",
        "financial_contracts.swap",
        "financial_contracts.swaption",
        "financial_contracts.term_deposit",
    ],
)
def test_financial_contracts_stream_pending_state(definition):
    def do_open(*args, **kwargs):
        # then
        assert stream.open_state is OpenState.Pending

    # given
    session = StubSession(is_open=True)
    stream = definition.get_stream(session=session)
    stream._stream._stream._do_open = do_open

    # when
    stream.open()



__all__ = (
    "Package",
    "view_actuals",
    "view_actuals_kpi",
    "view_summary",
    "view_summary_kpi",
)

from . import view_actuals, view_actuals_kpi, view_summary, view_summary_kpi
from ._enums import Package



from typing import List, Any, Union
from collections.abc import Iterable

import numpy as np
import pandas as pd


def try_copy_to_list(obj: Any) -> Union[List[Any], Any]:
    if isinstance(obj, Iterable) and not isinstance(obj, str):
        return list(obj)

    return obj


class Copier:
    @staticmethod
    def get_list(obj: Any) -> Union[List[Any], Any]:
        if isinstance(obj, Iterable) and not isinstance(obj, str):
            return list(obj)

        return [obj]


def convert_dict_to_df(data, columns) -> pd.DataFrame:
    if data:
        df = pd.DataFrame(np.array(data), columns=columns)
        if not df.empty:
            df.fillna(pd.NA, inplace=True)
            df = df.convert_dtypes()
    else:
        df = pd.DataFrame([], columns=columns)
    return df


def convert_content_data_to_df(
    content_data: dict, use_field_names_in_headers: bool = True
):
    if "headers" not in content_data or "data" not in content_data:
        return pd.DataFrame()

    if use_field_names_in_headers:
        key = "name"
    else:
        key = "title"

    headers_names = [header[key] for header in content_data["headers"]]
    return convert_dict_to_df(content_data["data"], headers_names)



from datetime import datetime

import refinitiv.data.content.ipa.financial_contracts as rdf


def repo_definition_01():
    definition = rdf.repo.Definition(
        start_date="2019-11-27",
        tenor="1M",
        underlying_instruments=[
            rdf.repo.UnderlyingContract(
                instrument_type="Bond",
                instrument_definition=rdf.bond.Definition(
                    instrument_code="US191450264="
                ),
            )
        ],
        pricing_parameters=rdf.repo.PricingParameters(
            market_data_date="2019-11-25",
        ),
        fields=["MarketDataDate", "ErrorMessage"],
    )

    return definition


def repo_definition_02():
    definition = rdf.repo.Definition(
        start_date="2020-08-17T00:00:00Z",
        end_date="2020-08-22T00:00:00Z",
        underlying_instruments=[
            rdf.repo.UnderlyingContract(
                instrument_type="Bond",
                instrument_definition=rdf.bond.Definition(
                    instrument_code="US191450264="
                ),
                pricing_parameters=rdf.repo.UnderlyingPricingParameters(
                    repo_parameters=rdf.repo.RepoParameters(initial_margin_percent=50)
                ),
            )
        ],
        pricing_parameters=rdf.repo.PricingParameters(
            market_data_date="2020-08-19T00:00:00Z"
        ),
    )

    return definition


def on_response(response, session, events_list=None):
    current_time = datetime.now().time()
    if events_list is not None:
        events_list.append("Response received for")
    print(current_time, "- Response received for", response.data.raw)


repo_universe = {
    "instrumentType": "Repo",
    "instrumentDefinition": {
        "InstrumentTag": "my tag",
        "startDate": "2020-08-17T00:00:00Z",
        "endDate": "2020-08-22T00:00:00Z",
        "underlyingInstruments": [
            {
                "instrumentType": "Bond",
                "instrumentDefinition": {"instrumentCode": "US191450264="},
                "pricingParameters": {"repoParameters": {"initialMarginPercent": 50}},
            }
        ],
    },
    "pricingParameters": {"marketDataDate": "2020-08-19T00:00:00Z"},
}



import refinitiv.data.content.ipa.financial_contracts as rdf
from refinitiv.data.content.ipa.curves import (
    zc_curves,
    forward_curves,
)
from refinitiv.data.content.ipa.financial_contracts import cap_floor, cds, option
from refinitiv.data.content.ipa.surfaces import cap

ZC_CURVE_DEFINITION = zc_curves.Definition(
    curve_definition=zc_curves.ZcCurveDefinitions(
        currency="CHF",
        name="CHF LIBOR Swap ZC Curve",
        discounting_tenor="OIS",
    ),
    curve_parameters=zc_curves.ZcCurveParameters(use_steps="True"),
)

ZC_CURVE_RESPONSE = {
    "error": {
        "id": "3a42afe0-28c1-4eca-a158-7a20944e5059",
        "code": "400",
        "message": "Validation error",
        "status": "Bad Request",
        "errors": [
            {
                "key": "zcCurvesRequest",
                "reason": 'zcCurvesRequest.universe.curveParameters.useSteps in body must be of type boolean: "string"',
            }
        ],
    }
}

FORWARD_CURVE_DEFINITION = forward_curves.Definition(
    curve_definition=forward_curves.SwapZcCurveDefinition(
        currency="EUR123",
        index_name="EURIBOR",
        name="EUR EURIBOR Swap ZC Curve",
        discounting_tenor="OIS",
    ),
    forward_curve_definitions=[
        forward_curves.ForwardCurveDefinition(
            index_tenor="3M",
            forward_curve_tag="ForwardTag",
            forward_start_date="2021-02-01",
            forward_curve_tenors=[
                "0D",
                "1D",
                "2D",
                "3M",
                "6M",
                "9M",
                "1Y",
                "2Y",
                "3Y",
                "4Y",
                "5Y",
                "6Y",
                "7Y",
                "8Y",
                "9Y",
                "10Y",
                "15Y",
                "20Y",
                "25Y",
            ],
        )
    ],
)

FORWARD_CURVE_RESPONSE = {
    "data": [
        {
            "error": {
                "id": "8abfe7bd-4b66-4b6d-b994-47d0758759c6/8abfe7bd-4b66-4b6d-b994-47d0758759c6",
                "code": "QPS-Curves.10",
                "message": "The service failed to find the curve constituents",
            }
        }
    ]
}

ZC_CURVES_DEFINITION = zc_curves.Definitions(
    [
        zc_curves.Definition(
            constituents={},
            curve_parameters=zc_curves.ZcCurveParameters(
                valuation_date="1719-08-21",
                price_side="Mid",
                interpolation_mode=zc_curves.ZcInterpolationMode.CUBIC_DISCOUNT,
            ),
            curve_definition=zc_curves.ZcCurveDefinitions(
                currency="EUR",
                index_name="EURIBOR",
                source="Refinitiv",
                discounting_tenor="OIS",
            ),
            curve_tag="TAG",
        ),
        zc_curves.Definition(
            curve_definition=zc_curves.ZcCurveDefinitions(
                currency="EUR",
                index_name="EURIBOR",
                source="Refinitiv",
                discounting_tenor="OIS",
            )
        ),
    ]
)

ZC_CURVES_RESPONSE = {
    "data": [
        {
            "curveTag": "TAG",
            "error": {
                "id": "d25267cf-bbfc-4230-84a4-ad8c33914e68/d25267cf-bbfc-4230-84a4-ad8c33914e68",
                "code": "QPS-Curves.7",
                "message": "The service failed to find the curve definition",
            },
        },
    ]
}

SURFACE_CAP_DEFINITION = cap.Definition(
    surface_tag="USD_Strike__Tenor_",
    underlying_definition=cap.CapSurfaceDefinition(
        instrument_code="USD123",  # <<<----
        discounting_type=cap.DiscountingType.OIS_DISCOUNTING,
    ),
    surface_layout=cap.SurfaceLayout(format=cap.Format.MATRIX),
    surface_parameters=cap.CapCalculationParams(
        valuation_date="2020-03-20",
        x_axis=cap.Axis.STRIKE,
        y_axis=cap.Axis.TENOR,
    ),
)

SURFACE_CAP_RESPONSE = {
    "data": [
        {
            "surfaceTag": "",
            "error": {
                "id": "5ccf67af-badb-47c3-8d74-c9bb9e7e994c/fc4a1fb4-1e3b-43cc-9eeb-84bed950a926",
                "status": "Error",
                "message": "The service failed to build the volatility surface",
                "code": "VolSurf.10300",
            },
        }
    ]
}

CAP_FLOOR_DEFINITION = cap_floor.Definition(
    instrument_tag="CapOnCms",
    stub_rule=rdf.cap_floor.StubRule.MATURITY,
    notional_ccy="USD123",
    start_date="2018-06-15",
    end_date="2022-06-15",
    notional_amount=1000000,
    index_name="Composite",
    index_tenor="5Y",
    interest_calculation_method="Dcb_Actual_360",
    interest_payment_frequency=rdf.cap_floor.Frequency.QUARTERLY,
    buy_sell=rdf.cap_floor.BuySell.BUY,
    cap_strike_percent=1,
    pricing_parameters=rdf.cap_floor.PricingParameters(
        skip_first_cap_floorlet=False, valuation_date="2020-02-07"
    ),
    fields=[
        "InstrumentTag",
        "InstrumentDescription",
        "StartDate",
        "EndDate",
        "InterestPaymentFrequency",
        "IndexRic",
        "CapStrikePercent",
        "FloorStrikePercent",
        "NotionalCcy",
        "NotionalAmount",
        "PremiumBp",
        "PremiumPercent",
        "MarketValueInDealCcy",
        "MarketValueInReportCcy",
        "ErrorMessage",
    ],
)

CAP_FLOOR_RESPONSE = {
    "headers": [
        {"type": "String", "name": "InstrumentTag"},
        {"type": "String", "name": "InstrumentDescription"},
        {"type": "DateTime", "name": "StartDate"},
        {"type": "DateTime", "name": "EndDate"},
        {"type": "String", "name": "InterestPaymentFrequency"},
        {"type": "String", "name": "IndexRic"},
        {"type": "Float", "name": "CapStrikePercent"},
        {"type": "Float", "name": "FloorStrikePercent"},
        {"type": "String", "name": "NotionalCcy"},
        {"type": "Float", "name": "NotionalAmount"},
        {"type": "Float", "name": "PremiumBp"},
        {"type": "Float", "name": "PremiumPercent"},
        {"type": "Float", "name": "MarketValueInDealCcy"},
        {"type": "Float", "name": "MarketValueInReportCcy"},
        {"type": "String", "name": "ErrorMessage"},
    ],
    "data": [
        [
            "CapOnCms",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "Market data error : an internal error occured. Failed to cast RateSurfaceInfoResponse.",
        ]
    ],
}

CDS_DEFINITION = cds.Definition(
    instrument_tag="Cds1_InstrumentCode",
    instrument_code="123BNPP5YEUAM=R",
    cds_convention=rdf.cds.CdsConvention.ISDA,
    end_date_moving_convention=rdf.cds.BusinessDayConvention.NO_MOVING,
    adjust_to_isda_end_date=True,
    pricing_parameters=rdf.cds.PricingParameters(market_data_date="2020-01-01"),
    fields=[
        "InstrumentTag",
        "ValuationDate",
        "InstrumentDescription",
        "StartDate",
        "EndDate",
        "SettlementDate",
        "UpfrontAmountInDealCcy",
        "CashAmountInDealCcy",
        "AccruedAmountInDealCcy",
        "AccruedBeginDate",
        "NextCouponDate",
        "UpfrontPercent",
        "ConventionalSpreadBp",
        "ParSpreadBp",
        "AccruedDays",
        "ErrorCode",
        "ErrorMessage",
    ],
)

CDS_RESPONSE = {
    "headers": [
        {"type": "String", "name": "InstrumentTag"},
        {"type": "DateTime", "name": "ValuationDate"},
        {"type": "String", "name": "InstrumentDescription"},
        {"type": "DateTime", "name": "StartDate"},
        {"type": "DateTime", "name": "EndDate"},
        {"type": "Date", "name": "SettlementDate"},
        {"type": "Float", "name": "UpfrontAmountInDealCcy"},
        {"type": "Float", "name": "CashAmountInDealCcy"},
        {"type": "Float", "name": "AccruedAmountInDealCcy"},
        {"type": "DateTime", "name": "AccruedBeginDate"},
        {"type": "Date", "name": "NextCouponDate"},
        {"type": "Float", "name": "UpfrontPercent"},
        {"type": "Float", "name": "ConventionalSpreadBp"},
        {"type": "Float", "name": "ParSpreadBp"},
        {"type": "Integer", "name": "AccruedDays"},
        {"type": "String", "name": "ErrorCode"},
        {"type": "String", "name": "ErrorMessage"},
    ],
    "data": [
        [
            "Cds1_InstrumentCode",
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            None,
            "QPS-DPS.1",
            "Technical error occured.",
        ]
    ],
}

OPTION_DEFINITION = option.Definition(
    instrument_code="123FCHI560000L1.p",
    underlying_type=rdf.option.UnderlyingType.ETI,
    fields=[
        "MarketValueInDealCcy",
        "DeltaPercent",
        "GammaPercent",
        "RhoPercent",
        "ThetaPercent",
        "VegaPercent",
        "ErrorCode",
        "ErrorMessage",
    ],
)

OPTION_RESPONSE = {
    "headers": [
        {"type": "Float", "name": "MarketValueInDealCcy"},
        {"type": "Float", "name": "DeltaPercent"},
        {"type": "Float", "name": "GammaPercent"},
        {"type": "Float", "name": "RhoPercent"},
        {"type": "Float", "name": "ThetaPercent"},
        {"type": "Float", "name": "VegaPercent"},
        {"type": "String", "name": "ErrorCode"},
        {"type": "String", "name": "ErrorMessage"},
    ],
    "data": [
        [
            None,
            None,
            None,
            None,
            None,
            None,
            "QPS-DPS.1",
            "Technical error occured. No Realtime Data retrieved by DAL. Instrument='FCHI560000L1.p' does not exist!",
        ]
    ],
}



# coding: utf8

from enum import Enum, unique


@unique
class ImpliedDepositDateConvention(Enum):
    FX_MARKET_CONVENTION = "FxMarketConvention"
    MONEY_MARKET_CONVENTION = "MoneyMarketConvention"



from typing import TYPE_CHECKING

from ...._curves._cross_currency_curves._definitions._search import (
    CrossCurrencyCurveGetDefinitionItem,
)
from ......_content_type import ContentType
from ......_tools import create_repr
from ....._content_provider_layer import ContentProviderLayer


if TYPE_CHECKING:
    from ...._curves._cross_currency_curves._definitions._types import (
        OptMainConstituentAssetClass,
        OptRiskType,
    )
    from ......_types import OptStr, OptBool, ExtendedParams


class Definition(ContentProviderLayer):
    """
    Returns the definitions of the available Commodities curves for the filters
    selected (e.g. sector, subSector...).

    Parameters
    ----------
    main_constituent_asset_class : MainConstituentAssetClass, optional
        The asset class used to generate the zero coupon curve. the possible values are:
        * fxforward   * swap
    risk_type : RiskType, optional
        The risk type to which the generated cross currency curve is sensitive. the
        possible value is:   * 'crosscurrency'
    base_currency : str, optional
        The base currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'eur').
    base_index_name : str, optional
        The name of the floating rate index (e.g., 'estr') applied to the base currency.
    curve_tag : str, optional
        A user-defined string to identify the interest rate curve. it can be used to
        link output results to the curve definition. limited to 40 characters. only
        alphabetic, numeric and '- _.#=@' characters are supported.
    id : str, optional
        The identifier of the cross currency definitions.
    is_non_deliverable : bool, optional
        An indicator whether the instrument is non-deliverable:   * true: the instrument
        is non-deliverable,   * false: the instrument is not non-deliverable. the
        property can be used to retrieve cross currency definition for the adjusted
        interest rate curve.
    name : str, optional
        The fxcross currency pair applied to the reference or pivot currency. it is
        expressed in iso 4217 alphabetical format (e.g., 'eur usd fxcross').
    quoted_currency : str, optional
        The quoted currency in the fxcross currency pair. it is expressed in iso 4217
        alphabetical format (e.g., 'usd').
    quoted_index_name : str, optional
        The name of the floating rate index (e.g., 'sofr') applied to the quoted
        currency.
    source : str, optional
        A user-defined string that is provided by the creator of a curve. curves created
        by refinitiv have the 'refinitiv' source.
    valuation_date : str, optional
        The date used to define a list of curves or a unique cross currency curve that
        can be priced at this date. the value is expressed in iso 8601 format:
        yyyy-mm-dd (e.g., '2021-01-01').
    extended_params : dict, optional
        If necessary other parameters.

    Examples
    --------
    >>> from refinitiv.data.content.ipa.curves._cross_currency_curves import definitions
    >>> definition = definitions.search.Definition(
    ...     base_currency="EUR",
    ...     quoted_currency="CHF"
    >>> )
    >>> response = definition.get_data()

    Using get_data_async
    >>> import asyncio
    >>> task = definition.get_data_async()
    >>> response = asyncio.run(task)
    """

    def __init__(
        self,
        main_constituent_asset_class: "OptMainConstituentAssetClass" = None,
        risk_type: "OptRiskType" = None,
        base_currency: "OptStr" = None,
        base_index_name: "OptStr" = None,
        curve_tag: "OptStr" = None,
        id: "OptStr" = None,
        is_non_deliverable: "OptBool" = None,
        name: "OptStr" = None,
        quoted_currency: "OptStr" = None,
        quoted_index_name: "OptStr" = None,
        source: "OptStr" = None,
        valuation_date: "OptStr" = None,
        extended_params: "ExtendedParams" = None,
    ) -> None:

        request_item = CrossCurrencyCurveGetDefinitionItem(
            main_constituent_asset_class=main_constituent_asset_class,
            risk_type=risk_type,
            base_currency=base_currency,
            base_index_name=base_index_name,
            curve_tag=curve_tag,
            id=id,
            is_non_deliverable=is_non_deliverable,
            name=name,
            quoted_currency=quoted_currency,
            quoted_index_name=quoted_index_name,
            source=source,
            valuation_date=valuation_date,
        )
        super().__init__(
            content_type=ContentType.CROSS_CURRENCY_CURVES_DEFINITIONS_SEARCH,
            universe=request_item,
            extended_params=extended_params,
        )

    def __repr__(self):
        return create_repr(
            self, middle_path="_cross_currency_curves.definitions.search"
        )



import os
import sqlite3
from logging import Logger
from typing import Optional, Iterable

import mysql.connector

from ._actions import Actions
from ._df_builder import bulk_build_df
from ._errors import (
    KeyNotFoundException,
    TableNotFoundException,
    UpdateDBException,
    DBModuleNotFoundException,
)
from ..._content_data import Data
from ..._df_builder import build_empty_df
from ...._core.log_reporter import _LogReporter
from ...._tools import parse_url, get_from_path, urljoin
from ...._tools.templates import OldStyleStringTemplate as StringTemplate


def has_table(table_name, names):
    return table_name in names or table_name.lower() in names


def uri_to_path(value):
    if isinstance(value, str):
        p = parse_url(value)
    else:
        p = value
    path = p.path
    path = urljoin(p.netloc, path)
    return os.path.normpath(path)


def get_query(value):
    p = parse_url(value)
    if p.scheme == "file":
        path = uri_to_path(p)
        with open(path, "r") as f:
            result = f.read()
    else:
        result = value
    return result


class StringTemplateWithDottedNames(StringTemplate):
    # The same pattern as original, just the dot added
    idpattern = r"(?a:[_a-z][_a-z0-9\.]*)"


def prepare_insert_query(query: str, data: dict, file_attrs: dict = None) -> str:
    """
    Parameters
    ----------
    query: str
        query is like
        "INSERT (INSERT_DATE, INSTRUMENT) VALUES (#File.DateTime#, #Fields.Instrument#)"
    data: dict
    file_attrs: dict, optional

    Returns
    -------
    str
    """
    tpl = StringTemplateWithDottedNames(query)

    keys = {}
    for name in tpl.names():
        if name.startswith("Fields.") or name.startswith("File."):
            value = get_from_path(
                data if name.startswith("Fields.") else file_attrs,
                path=name.split(".", maxsplit=1)[1],
            )
            # if value is str like "Cabela's LLC"
            # then it will insert in query like:
            #   "INSERT (organization, score) VALUES ('Cabela's LLC', '0.123')"
            # and will get error while db.exec(query)
            # This 'if' statement is fixing the situation above
            if "Name" in name and isinstance(value, str) and "'" in value:
                value = value.replace("'", r"''")
            keys[name] = "null" if value is None else value
        else:
            raise KeyNotFoundException()

    return tpl.substitute(keys)


def prepare_search_query(
    query: str, universe: Iterable[str], columns: Iterable[str] = None
) -> str:
    """
    Parameters
    ----------
    query: str
        query is like:
         "SELECT * FROM TABLE WHERE INSTRUMENT IN #universe#"
         "SELECT COLUMN1, COLUMN2 FROM TABLE WHERE INSTRUMENT IN #universe#"
    universe: list of str
        universe is like ['4295865175', '4295906529']
    columns: list of str, optional
        columns is like ['COLUMN1', 'COLUMN2']

    Returns
    -------
    str
        Returns query is like
        "SELECT COLUMN1, COLUMN2
         FROM TABLE WHERE INSTRUMENT IN ('4295865175', '4295906529')"
    """
    if "*" in query and columns:
        query = query.replace("*", ", ".join(columns))

    return StringTemplateWithDottedNames(query).safe_substitute(
        universe=f"({', '.join(map(repr, universe))})"
    )


class DBManager(_LogReporter):
    def __init__(self, connection, config, actions: Actions, logger) -> None:
        super().__init__(logger=logger)
        self._connection = connection
        self._cursor = self._connection.cursor()
        self._config = config
        self._actions = actions

    def exec(self, sql):
        self._cursor.execute(sql)
        result = self._cursor.fetchall()
        return result

    def commit(self):
        self._connection.commit()

    def close(self):
        self._cursor.close()
        self._connection.close()

    def has_table(self, table_name, check_if_exists_query):
        has = False
        if check_if_exists_query:
            result = self.exec(check_if_exists_query)
            has = table_name in result or table_name.lower() in result

        return has

    def create_table(self):
        self._actions.update()
        created_tables = self._actions.get_created_tables()
        items = self._config.get("create-table-queries")

        for item in items:
            table_name = item.get("table-name")

            if table_name in created_tables:
                continue

            check_if_exists_query = item.get("check-if-exists-query")
            if self.has_table(table_name, check_if_exists_query):
                continue

            create_query = get_query(item.get("create-query"))
            if not create_query:
                raise TableNotFoundException("table doesn't exists")

            self._info(f'executing the "CREATE TABLE" query, table_name={table_name} ')
            self.exec(create_query)
            self.commit()
            self._actions.table_created(**{"table-name": table_name})

    def update_db(self, filesdata_for_update: list):
        if not filesdata_for_update and not self._actions.get_downloaded():
            message = "no init file found in folder"
            self._actions.add("ERROR", message=message)
            self._error(f"error: {message}")
            raise FileNotFoundError(message)

        if not filesdata_for_update:
            self._actions.add("NO DATA TO UPDATE")
            self._info("no data to update")
            return

        self.create_table()
        self.__update_db(filesdata_for_update)

    def __update_db(self, filesdata_for_update: list):
        insert_queries = list(map(get_query, self._config.get("insert-queries")))
        stop_on_error = self._config.get("stop-on-error")

        for filename, items, file_attrs in filesdata_for_update:
            lines = 0
            for line_n, data in enumerate(items):
                for query in insert_queries:
                    try:
                        query = prepare_insert_query(query, data, file_attrs)
                        self.exec(query)
                    except Exception as e:
                        error_details = {
                            "message": str(e),
                            "line_number": line_n,
                            "filename": filename,
                        }
                        self._actions.add("ERROR", **error_details)
                        error_details["query"] = query
                        self._info(f"error: {error_details}")

                        if stop_on_error:
                            raise UpdateDBException(str(error_details))
                        continue

                lines += 1

            self.commit()
            self._actions.updated(filename=filename, lines=lines)

    def cleanup_db(self):
        clean_up_queries = self._config.get("clean-up-queries")
        for query in clean_up_queries:
            self.exec(query)
        self.commit()
        self._actions.cleaned_up_db()
        self._info("clean up database")

    def get_data(self, universe: Iterable[str], fields: Iterable[str] = None) -> Data:
        fields = fields or []
        query = self._config.get("search-query")

        column_by_field = self._config["output-fields-mapping"].as_dict()
        field_by_column = {v: k for k, v in column_by_field.items()}
        columns = [
            field_by_column[field] for field in fields if field in field_by_column
        ]

        try:
            query = prepare_search_query(query, universe, columns)
            raw = self.exec(query)
            error = None
        except Exception as e:
            raw = {}
            error = {
                "message": str(e),
            }
            self._actions.add("ERROR", **error)
            error["query"] = query
            self._info(f"error: {error}")

        if error:
            return Data(raw, build_empty_df)

        return Data(
            raw,
            bulk_build_df,
            column_by_field=column_by_field,
            columns=self._cursor.description,
        )


def create_sqlite_connector(config):
    connection = sqlite3.connect(
        database=config.get("connection.parameters.database"), uri=True
    )
    return connection


def create_mysql_connector(config):
    conn_params = config["connection"]["parameters"]
    connection = mysql.connector.connect(
        user=conn_params.get("user"),
        password=conn_params.get("password"),
        host=conn_params.get("host"),
        port=conn_params.get("port"),
        database=conn_params.get("database"),
    )
    return connection


connections_makers_by_db_modules = {
    "sqlite3": create_sqlite_connector,
    "mysql": create_mysql_connector,
}


def create_db_manager(config, actions, logger):
    db_module_name = config.get("connection.module")
    create_connection = connections_makers_by_db_modules.get(db_module_name)

    if not create_connection:
        raise DBModuleNotFoundException(f"module {db_module_name} not found")

    connection = create_connection(config)
    manager = DBManager(connection, config, actions, logger)
    return manager


def create_db_manager_by_package_name(
    package_name,
    actions: Optional[Actions] = None,
    logger: Optional[Logger] = None,
):
    from ._package_manager import create_logpath, get_console_logger
    from .... import _configure

    id_ = f"{package_name}"
    config = _configure.get(f"bulk.{package_name}.db")
    actions = actions or Actions(
        id_=id_,
        logpath=create_logpath(package_name),
    )
    logger = logger or get_console_logger(id_)
    manager = create_db_manager(config, actions, logger)
    return manager



import pytest
import refinitiv.data as rd
from tests.unit.conftest import StubSession


def test_stream_is_open_raise_error(stream):
    # then
    with pytest.raises(AttributeError, match="object has no attribute 'is_open'"):
        # when
        stream.is_open


def test_stream_is_close_raise_error(stream):
    # then
    with pytest.raises(AttributeError, match="object has no attribute 'is_close'"):
        # when
        stream.is_close


def test_stream_is_pause_raise_error(stream):
    # then
    with pytest.raises(AttributeError, match="object has no attribute 'is_pause'"):
        # when
        stream.is_pause


def test_stream_is_paused_raise_error(stream):
    # then
    with pytest.raises(AttributeError, match="object has no attribute 'is_paused'"):
        # when
        stream.is_paused


@pytest.mark.parametrize(
    "method",
    [
        "open",
        "open_async",
        "close",
    ],
)
def test_stream_has_methods(method, stream):
    # when
    has = hasattr(stream, method)
    attr = getattr(stream, method)
    has = has and callable(attr)

    # then
    assert has is True


@pytest.mark.parametrize(
    "method",
    [
        "close_async",
    ],
)
def test_stream_doesnt_have_methods(method, stream):
    # when
    not_has = not hasattr(stream, method)

    # then
    assert not_has is True


@pytest.mark.parametrize(
    "expected_repr, defn_with_stream",
    [
        (
            "<refinitiv.data.content.pricing.Stream object at",
            rd.content.pricing.Definition("universe"),
        ),
        (
            "<refinitiv.data.content.trade_data_service.Stream object at",
            rd.content.trade_data_service.Definition(),
        ),
        (
            "<refinitiv.data.delivery.rdp_stream.RDPStream object at",
            rd.delivery.rdp_stream.Definition("", [], [], {}, ""),
        ),
        (
            "<refinitiv.data.delivery.omm_stream.OMMStream object at",
            rd.delivery.omm_stream.Definition(""),
        ),
        (
            "<refinitiv.data.content.pricing.chain.Stream object at",
            rd.content.pricing.chain.Definition("universe"),
        ),
        (
            "<refinitiv.data.content.ipa.financial_contracts.Stream object at",
            rd.content.ipa.financial_contracts.bond.Definition("universe"),
        ),
    ],
    ids=[
        "pricing",
        "trade_data_service",
        "rdp_stream",
        "omm_stream",
        "pricing.chain",
        "financial_contracts.bond",
    ],
)
def test_stream_repr(expected_repr, defn_with_stream):
    # given
    session = StubSession(is_open=True)
    stream = defn_with_stream.get_stream(session)

    # when
    testing_repr = repr(stream)

    # then
    assert expected_repr in testing_repr, testing_repr



import time

from refinitiv.data.content.pricing.chain._stream import StreamingChain
from tests.unit.conftest import StubSession
from ....conftest import TIMEOUT


def test_on_complete_called_once():
    on_complete_called = 0

    def called_once(*args):
        nonlocal on_complete_called
        on_complete_called += 1

    session = StubSession(is_open=True)
    streaming_chain = StreamingChain("test", session=session)
    streaming_chain.on_complete(called_once)

    streaming_chain._chain_records._on_stream_complete("")
    streaming_chain._chain_records._on_stream_complete("")
    streaming_chain._chain_records._on_stream_complete("")
    streaming_chain._on_stream_complete(streaming_chain)

    time.sleep(TIMEOUT)
    assert on_complete_called == 1


def test_streaming_chain_custom_service():
    # given
    custom_service = "TEST_SERVICE"
    session = StubSession(is_open=True)

    # when
    streaming_chain = StreamingChain("test", session=session, service=custom_service)

    # then
    assert streaming_chain._service == custom_service
    assert streaming_chain._chain_records.service == custom_service



from typing import Optional

from ._turn_adjustment import TurnAdjustment
from .._enums import InterpolationMode
from ...._enums._extrapolation_mode import ExtrapolationMode
from ...._object_definition import ObjectDefinition


class CrossCurrencyCurveParameters(ObjectDefinition):
    """
    Creates the Cross Currency curve definition with the definition provided.

    Parameters
    ----------
    extrapolation_mode : ExtrapolationMode, optional

    interpolation_mode : InterpolationMode, optional

    turn_adjustment : TurnAdjustment, optional

    """

    def __init__(
        self,
        extrapolation_mode: Optional[ExtrapolationMode] = None,
        interpolation_mode: Optional[InterpolationMode] = None,
        turn_adjustment: Optional[TurnAdjustment] = None,
    ) -> None:
        super().__init__()
        self.extrapolation_mode = extrapolation_mode
        self.interpolation_mode = interpolation_mode
        self.turn_adjustment = turn_adjustment

    @property
    def extrapolation_mode(self):
        """
        :return: enum ExtrapolationMode
        """
        return self._get_enum_parameter(ExtrapolationMode, "extrapolationMode")

    @extrapolation_mode.setter
    def extrapolation_mode(self, value):
        self._set_enum_parameter(ExtrapolationMode, "extrapolationMode", value)

    @property
    def interpolation_mode(self):
        """
        :return: enum InterpolationMode
        """
        return self._get_enum_parameter(InterpolationMode, "interpolationMode")

    @interpolation_mode.setter
    def interpolation_mode(self, value):
        self._set_enum_parameter(InterpolationMode, "interpolationMode", value)

    @property
    def turn_adjustment(self):
        """
        :return: object TurnAdjustment
        """
        return self._get_object_parameter(TurnAdjustment, "turnAdjustment")

    @turn_adjustment.setter
    def turn_adjustment(self, value):
        self._set_object_parameter(TurnAdjustment, "turnAdjustment", value)



from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    ShiftDefinition,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    ButterflyShift,
    CombinedShift,
    FlatteningShift,
    LongEndShift,
    ParallelShift,
    ShortEndShift,
    TimeBucketShift,
    TwistShift,
)


def test_shift_definition():
    # given
    expected_dict = {
        "butterflyShift": {},
        "combinedShifts": [{}],
        "flatteningShift": {},
        "longEndShift": {},
        "parallelShift": {},
        "shortEndShift": {},
        "timeBucketShifts": [{}],
        "twistShift": {},
    }

    # when
    testing_obj = ShiftDefinition(
        butterfly_shift=ButterflyShift(),
        combined_shifts=[CombinedShift()],
        flattening_shift=FlatteningShift(),
        long_end_shift=LongEndShift(),
        parallel_shift=ParallelShift(),
        short_end_shift=ShortEndShift(),
        time_bucket_shifts=[TimeBucketShift()],
        twist_shift=TwistShift(),
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert testing_obj.butterfly_shift == ButterflyShift()
    assert testing_obj.combined_shifts == [CombinedShift()]
    assert testing_obj.flattening_shift == FlatteningShift()
    assert testing_obj.long_end_shift == LongEndShift()
    assert testing_obj.parallel_shift == ParallelShift()
    assert testing_obj.short_end_shift == ShortEndShift()
    assert testing_obj.time_bucket_shifts == [TimeBucketShift()]
    assert testing_obj.twist_shift == TwistShift()



# coding: utf8


from enum import Enum, unique


@unique
class VolatilityAdjustmentType(Enum):
    CONSTANT_CAP = "ConstantCap"
    CONSTANT_CAPLET = "ConstantCaplet"
    NORMALIZED_CAP = "NormalizedCap"
    NORMALIZED_CAPLET = "NormalizedCaplet"
    PB_UNDEFINED = "PbUndefined"
    SHIFTED_CAP = "ShiftedCap"



from refinitiv.data.content.ipa._curves._cross_currency_curves._curves import (
    CrossCurrencyCurveDefinitionPricing,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._curves._enums import (
    ConstituentOverrideMode,
)
from refinitiv.data.content.ipa._curves._cross_currency_curves._enums import (
    MainConstituentAssetClass,
    RiskType,
)


def test_cross_currency_curve_definition_pricing():
    # given
    expected_dict = {
        "constituentOverrideMode": "MergeWithDefinition",
        "mainConstituentAssetClass": "FxForward",
        "riskType": "CrossCurrency",
        "baseCurrency": "string",
        "baseIndexName": "string",
        "id": "string",
        "ignoreExistingDefinition": True,
        "isNonDeliverable": True,
        "name": "string",
        "quotedCurrency": "string",
        "quotedIndexName": "string",
        "source": "string",
    }

    # when
    testing_obj = CrossCurrencyCurveDefinitionPricing(
        constituent_override_mode=ConstituentOverrideMode.MERGE_WITH_DEFINITION,
        main_constituent_asset_class=MainConstituentAssetClass.FX_FORWARD,
        risk_type=RiskType.CROSS_CURRENCY,
        base_currency="string",
        base_index_name="string",
        id="string",
        ignore_existing_definition=True,
        is_non_deliverable=True,
        name="string",
        quoted_currency="string",
        quoted_index_name="string",
        source="string",
    )

    # then
    assert testing_obj.get_dict() == expected_dict

    assert (
        testing_obj.constituent_override_mode
        == ConstituentOverrideMode.MERGE_WITH_DEFINITION
    )
    assert (
        testing_obj.main_constituent_asset_class == MainConstituentAssetClass.FX_FORWARD
    )
    assert testing_obj.risk_type == RiskType.CROSS_CURRENCY
    assert testing_obj.base_currency == "string"
    assert testing_obj.base_index_name == "string"
    assert testing_obj.id == "string"
    assert testing_obj.ignore_existing_definition == True
    assert testing_obj.is_non_deliverable == True
    assert testing_obj.name == "string"
    assert testing_obj.quoted_currency == "string"
    assert testing_obj.quoted_index_name == "string"
    assert testing_obj.source == "string"



from typing import Any, Callable, List, Optional, TYPE_CHECKING, Union

from ._custom_instruments_data_provider import (
    is_instrument_id,
    get_valid_symbol,
    get_valid_symbol_request,
    symbol_with_user_id,
    get_user_id,
)
from .._universe_stream import _UniverseStream
from .._universe_streams import UniverseStreamFacade, _UniverseStreams
from ..._content_type import ContentType
from ..._core.session import get_valid_session
from ..._tools import (
    PRICING_DATETIME_PATTERN,
    cached_property,
    create_repr,
    make_callback,
)
from ..._tools._dataframe import convert_df_columns_to_datetime_re
from ...delivery._data._data_provider import DataProviderLayer
from ...delivery._data._endpoint_data import RequestMethod
from ...delivery._stream.base_stream import StreamOpenWithUpdatesMixin

if TYPE_CHECKING:
    from ..._types import OptStr
    import pandas
    from ..._core.session import Session


def uuid_from_symbol(symbol: str) -> str:
    return symbol.rsplit(".", 1)[-1]


def _init_universe(universe, session, uuid=None):
    data_provider_layer = DataProviderLayer(
        data_type=ContentType.CUSTOM_INSTRUMENTS_INSTRUMENTS,
        universe=universe,
    )

    result = []
    for item in universe:
        # a20140be-3648-4892-9d1b-ce78ee8617fd
        if is_instrument_id.match(item):
            instrument_response = data_provider_layer.get_data(
                session, method=RequestMethod.GET
            )
            # S)MyNewInstrument5.GE-1525-0
            symbol = instrument_response.data.raw.get("symbol")

        else:
            # S)MyNewInstrument5
            if not symbol_with_user_id.match(item) and not uuid:
                # S)MyNewInstrument5.GE-1525-0
                symbol = get_valid_symbol_request(item, session)
                # GE-1525-0
                uuid = uuid_from_symbol(symbol)

            else:
                symbol = get_valid_symbol(item, uuid)

        result.append(symbol)

    return result


class CustomInstrumentsStream(UniverseStreamFacade):
    pass


class CustomInstsUniverseStreams(_UniverseStreams):
    def _create_stream_by_name(self, name):
        return CustomInstsUniverseStream(
            content_type=self._content_type,
            session=self._session,
            name=name,
            fields=self.fields,
            service=self._service,
            on_refresh=self._on_stream_refresh,
            on_status=self._on_stream_status,
            on_update=self._on_stream_update,
            on_complete=self._on_stream_complete,
            on_error=self._on_stream_error,
            api=self._api,
            extended_params=self._extended_params,
            parent_id=self._id,
        )


class CustomInstsUniverseStream(_UniverseStream):
    def _filter_fields(self, fields):
        return {k: v for k, v in fields.items() if k in self._fields}


class Stream(StreamOpenWithUpdatesMixin):
    def __init__(
        self,
        session: "Session" = None,
        universe: Union[str, List[str]] = None,
        fields: Optional[list] = None,
        service: "OptStr" = None,
        api: "OptStr" = None,
        extended_params: Optional[dict] = None,
        uuid: "OptStr" = None,
    ) -> None:
        self._session = get_valid_session(session)
        self._always_use_default_session = session is None
        self._universe = universe or []
        self._fields = fields
        self._service = service
        self._api = api
        self._extended_params = extended_params

        if not uuid:
            uuid = get_user_id(self._session)
        self._uuid = uuid

    @cached_property
    def _stream(self) -> _UniverseStreams:
        if self._api:
            content_type = ContentType.STREAMING_CUSTOM
        else:
            content_type = ContentType.STREAMING_CUSTOM_INSTRUMENTS

        return CustomInstsUniverseStreams(
            content_type=content_type,
            item_facade_class=CustomInstrumentsStream,
            universe=self._universe,
            session=self._session,
            fields=self._fields,
            service=self._service,
            api=self._api,
            extended_params=self._extended_params,
        )

    def _get_fields(self, universe: str, fields: Optional[list] = None) -> dict:
        _fields = {
            universe: {
                key: value
                for key, value in self._stream[universe].items()
                if fields is None or key in fields
            }
        }
        return _fields

    def get_snapshot(
        self,
        universe: Union[str, List[str], None] = None,
        fields: Optional[List[str]] = None,
        convert: bool = True,
    ) -> "pandas.DataFrame":
        if isinstance(universe, str):
            universe = [universe]
        if universe is not None:
            universe = [get_valid_symbol(i, self._uuid) for i in universe]
        df = self._stream.get_snapshot(
            universe=universe, fields=fields, convert=convert
        )
        convert_df_columns_to_datetime_re(df, PRICING_DATETIME_PATTERN)
        return df

    def on_refresh(self, func: Callable[[Any, str, "Stream"], Any]) -> "Stream":
        self._stream.on_refresh(make_callback(func))
        return self

    def on_update(self, func: Callable[[Any, str, "Stream"], Any]) -> "Stream":
        self._stream.on_update(make_callback(func))
        return self

    def on_status(self, func: Callable[[Any, str, "Stream"], Any]) -> "Stream":
        self._stream.on_status(make_callback(func))
        return self

    def on_complete(self, func: Callable[["Stream"], Any]) -> "Stream":
        self._stream.on_complete(func)
        return self

    def on_error(self, func: Callable) -> "Stream":
        self._stream.on_error(make_callback(func))
        return self

    def add_fields(self, fields):
        """
        Add fields to the fields list.

        Parameters
        ----------
        fields: str, list of str, optional
            List of fields to add.

        Examples
        --------
        >>> from refinitiv.data.content import custom_instruments
        >>> definition = custom_instruments.Definition(
        ...    ["Instrument1", "Instrument2", "Instrument3"],
        ...    fields=["TRDPRC_1", "ASK"]
        ...)
        >>> stream = definition.get_stream()
        >>> stream.add_fields("BID")
        """
        self._stream.add_fields(fields)

    def remove_fields(self, fields):
        """
        Remove fields from the fields list.

        Parameters
        ----------
        fields: str, list of str, optional
            List of fields to remove.

        Examples
        --------
        >>> from refinitiv.data.content import custom_instruments
        >>> definition = custom_instruments.Definition(
        ...    ["Instrument1", "Instrument2", "Instrument3"],
        ...    fields=["TRDPRC_1", "ASK"]
        ...)
        >>> stream = definition.get_stream()
        >>> stream.remove_fields("ASK")
        """
        self._stream.remove_fields(fields)

    def add_instruments(self, instruments):
        """
        Add instruments to the stream universe.

        Parameters
        ----------
        instruments: str, list of str, optional
            List of instruments to add.

        Examples
        --------
        >>> from refinitiv.data.content import custom_instruments
        >>> definition = custom_instruments.Definition(
        ...    ["Instrument1", "Instrument2", "Instrument3"],
        ...)
        >>> stream = definition.get_stream()
        >>> stream.add_instruments("Instrument4")
        """
        instruments = _init_universe(instruments, self._session)
        self._stream.add_instruments(instruments)

    def remove_instruments(self, instruments):
        """
        Remove instruments from the stream universe.

        Parameters
        ----------
        instruments: str, list of str, optional
            List of instruments to remove.

        Examples
        --------
        >>> from refinitiv.data.content import custom_instruments
        >>> definition = custom_instruments.Definition(
        ...    ["Instrument1", "Instrument2", "Instrument3"],
        ...)
        >>> stream = definition.get_stream()
        >>> stream.remove_instruments("Instrument2")
        """
        instruments = _init_universe(instruments, self._session)
        self._stream.remove_instruments(instruments)

    def __iter__(self):
        return self._stream.__iter__()

    def __getitem__(self, item) -> "UniverseStreamFacade":
        item = get_valid_symbol(item, self._uuid)
        return self._stream.__getitem__(item)

    def __len__(self) -> int:
        return self._stream.__len__()

    def __repr__(self):
        return create_repr(
            self,
            class_name=self.__class__.__name__,
            content=f"{{name='{self._universe}'}}",
        )



from ..._models import BidAskMid  # noqa
from ..._models import InterpolationWeight  # noqa
from ._surface_output import *
from ._volatility_surface_point import *
from ._surface import *
from ._moneyness_weight import *
from ._surface_filters import *
from ._strike_filter import *
from ._strike_filter_range import *
from ._maturity_filter import *



from refinitiv.data.delivery._data._endpoint_data_provider import EndpointRequestFactory


def test_endpoint_request_factory():
    # given
    factory = EndpointRequestFactory()
    body_params = {"key": "value"}
    expected_value = body_params

    # when
    testing_value = factory.extend_body_parameters(body_params)

    # then
    assert expected_value == testing_value



import json

import pytest

from refinitiv.data import _configure as configure
from refinitiv.data._errors import RDError


def test_merge_two_files_top_more_complex(tmp_path):
    p = tmp_path / "top.json"
    d = {
        "prop": {
            "foo": "rewritten",
            "inner_prop": {
                "inner_foo": "rewritten",
            },
        }
    }
    p.write_text(json.dumps(d))
    paths = [p]
    p = tmp_path / "bottom.json"
    d = {
        "prop": {
            "foo": "value",
            "buzz": "buzz_value",
            "inner_prop": {
                "inner_foo": "value",
                "inner_buzz": "buzz_value",
            },
        }
    }
    p.write_text(json.dumps(d))
    paths.append(p)

    configure.reload()
    cfg = configure._create_rdpconfig(paths)
    assert cfg["prop"]["foo"] == "rewritten"
    assert cfg["prop"]["inner_prop"]["inner_foo"] == "rewritten"
    assert cfg["prop"]["buzz"] == "buzz_value"
    assert "inner_buzz" in cfg.get("prop.inner_prop"), str(cfg)
    assert cfg.get("prop.inner_prop.inner_buzz") == "buzz_value", str(cfg)


def test_merge_two_files_top_more_simple(tmp_path):
    p = tmp_path / "top.json"
    p.write_text('{"prop_1": "value_1", "prop_2": "value_2"}')
    paths = [p]
    p = tmp_path / "bottom.json"
    p.write_text('{"prop": "value"}')
    paths.append(p)

    configure.reload()
    cfg = configure._create_rdpconfig(paths)
    assert cfg.get("prop") == "value"
    assert cfg.get("prop_2") == "value_2"
    assert cfg.get("prop_1") == "value_1"


def test_merge_two_files_bottom_more(tmp_path):
    p = tmp_path / "top.json"
    p.write_text('{"prop": "prop_value"}')
    paths = [p]
    p = tmp_path / "bottom.json"
    p.write_text('{"prop_1": "value_1", "prop_2": "value_2"}')
    paths.append(p)

    configure.reload()
    cfg = configure._create_rdpconfig(paths)
    assert cfg.get("prop_1") == "value_1"
    assert cfg.get("prop_2") == "value_2"
    assert cfg.get("prop") == "prop_value"


def test_merge_two_files_equal(tmp_path):
    p = tmp_path / "top.json"
    p.write_text('{"prop": "value_2"}')
    paths = [p]
    p = tmp_path / "bottom.json"
    p.write_text('{"prop": "value_1"}')
    paths.append(p)

    configure.reload()
    cfg = configure._create_rdpconfig(paths)
    assert cfg.get("prop") == "value_2"


def test_merge_three_files(tmp_path):
    p = tmp_path / "top.json"
    p.write_text('{"prop": "value_3"}')
    paths = [p]
    p = tmp_path / "middle.json"
    p.write_text('{"prop": "value_2"}')
    paths.append(p)
    p = tmp_path / "bottom.json"
    p.write_text('{"prop": "value_1"}')
    paths.append(p)

    configure.reload()
    cfg = configure._create_rdpconfig(paths)
    assert cfg.get("prop") == "value_3"


def test_merge_config_with_config_from_file(tmp_path):
    p = tmp_path / "config.json"
    p.write_text('{"prop": "value_2"}')
    paths = [p]

    configure.reload()
    cfg = configure._create_rdpconfig(paths)
    cfg.update(
        {
            "prop": "value_1",
            "prop_1": "value_11",
        }
    )
    assert "value_1" == cfg.get("prop")
    assert "value_11" == cfg.get("prop_1")


def test_exception_on_bad_json(tmp_path):
    cfg_path = tmp_path / "config.json"
    cfg_path.write_text('{"prop": "value",}')

    configure.reload()

    with pytest.raises(RDError) as exc:
        configure._read_config_file(cfg_path)
        assert str(cfg_path) in exc.text


def test_no_error_while_no_file(tmp_path):
    configure.reload()
    configure._create_rdpconfig([tmp_path / "config"])



from ._eti_barrier_definition import EtiBarrierDefinition
from ._eti_binary_definition import EtiBinaryDefinition
from ._eti_cbbc_definition import EtiCbbcDefinition
from ._eti_definition import EtiDefinition
from ._eti_double_barriers_definition import EtiDoubleBarriersDefinition
from ._eti_fixing_info import EtiFixingInfo
from ._eti_underlying_definition import EtiUnderlyingDefinition



from typing import List, Dict

from dateutil.parser import parse

from ._context import RDPMixin
from ._cust_inst_context import CustInstContext


class CustInstRDPContext(RDPMixin, CustInstContext):
    def get_headers(self, headers: dict) -> List[str]:
        return [
            item["name"].lower() if item["name"] == "DATE" else item["name"]
            for item in headers
        ]

    def _parse_list_to_data(self, raw: list, field_to_idx: dict) -> List[List[dict]]:
        data = []

        for raw_item in raw:
            if not raw_item:
                continue

            ric = raw_item["universe"]["ric"]
            headers = self.get_headers(raw_item["headers"])

            for raw_item_data in raw_item["data"]:
                template = {"instrument": ric}

                for header, raw_data in zip(headers, raw_item_data):
                    try:
                        parse(raw_data, fuzzy=False)
                        raw_data = f"{raw_data} 00:00:00"
                        template.update({header: raw_data})
                    except (ValueError, TypeError):
                        template.update({header: raw_data})

                item = []
                for field in field_to_idx:
                    item.insert(field_to_idx[field], template.get(field))

                data.append(item)

        return data

    def _parse_dict_to_data(self, raw: dict, field_to_idx: dict) -> List[List[dict]]:
        data = []
        ric = raw["universe"]["ric"]
        headers = self.get_headers(raw["headers"])

        for raw_data_item in raw["data"]:
            template = {"instrument": ric}

            for header, raw_data in zip(headers, raw_data_item):
                try:
                    parse(raw_data, fuzzy=False)
                    raw_data = f"{raw_data} 00:00:00"
                    template.update({header: raw_data})
                except (ValueError, TypeError):
                    template.update({header: raw_data})

            item = []
            for field in field_to_idx:
                item.insert(field_to_idx[field], template.get(field))

            data.append(item)

        return data

    def prepare_data(self, raw, fields, *args) -> list:
        field_to_idx = {"instrument": 0, "date": 1}
        field_to_idx.update({field: fields.index(field) + 2 for field in fields})

        if isinstance(raw, dict):
            return self._parse_dict_to_data(raw, field_to_idx)

        elif isinstance(raw, list):
            return self._parse_list_to_data(raw, field_to_idx)

    def prepare_headers(self, raw, *args) -> List[Dict[str, str]]:
        headers = [
            {"name": "instrument", "title": "Instrument"},
            {"name": "date", "title": "Date"},
        ]

        for item in raw:
            headers.append({"name": item, "title": item})

        return headers
